# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for vizier.pyvizier.oss.study_config."""
import datetime

from vizier._src.service import constants
from vizier._src.service import key_value_pb2
from vizier._src.service import study_pb2
from vizier.service import pyvizier as vz

from google.protobuf import struct_pb2
from vizier._src.pyvizier.oss import compare
from absl.testing import absltest
from absl.testing import parameterized


class StudyConfigTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.pconfigs = [
        study_pb2.StudySpec.ParameterSpec(
            parameter_id='learning_rate',
            double_value_spec=study_pb2.StudySpec.ParameterSpec.DoubleValueSpec(
                min_value=0.00001, max_value=1.0),
            scale_type=study_pb2.StudySpec.ParameterSpec.ScaleType
            .UNIT_LINEAR_SCALE),
        study_pb2.StudySpec.ParameterSpec(
            parameter_id='optimizer',
            categorical_value_spec=study_pb2.StudySpec.ParameterSpec
            .CategoricalValueSpec(values=['adagrad', 'adam', 'experimental'])),
    ]

  def testCreationFromAndToProtoStudy(self):
    expected_automated_stopping_config = (
        study_pb2.StudySpec.DefaultEarlyStoppingSpec())

    study_config_proto = study_pb2.StudySpec(
        algorithm='QUASI_RANDOM_SEARCH',
        metrics=[
            study_pb2.StudySpec.MetricSpec(
                metric_id='pr-auc',
                goal=study_pb2.StudySpec.MetricSpec.GoalType.MAXIMIZE,
            )
        ],
        default_stopping_spec=expected_automated_stopping_config,
        metadata=[
            key_value_pb2.KeyValue(
                key='foo',
                ns=vz.Namespace(['ns_bar']).encode(),
                value='val',
            )
        ],
    )

    study_config_proto.parameters.extend(self.pconfigs)
    # Test all proprties.
    sc = vz.StudyConfig.from_proto(study_config_proto)
    expected = vz.MetricsConfig(
        [
            vz.MetricInformation(
                name='pr-auc', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    self.assertEqual(sc.metric_information, expected)
    self.assertEqual(sc.algorithm, 'QUASI_RANDOM_SEARCH')
    self.assertIsNone(sc.pythia_endpoint)
    self.assertEqual(sc.single_objective_metric_name, 'pr-auc')
    self.assertTrue(sc.is_single_objective)

    compare.assertProto2Equal(self, expected_automated_stopping_config,
                              sc.automated_stopping_config.to_proto())
    compare.assertProto2Equal(self, study_config_proto, sc.to_proto())
    _ = vz.StudyConfig.from_problem(sc.to_problem())  # smoke test.

  @parameterized.parameters([
      ('my custom algorithm'),
      ('QUASI_RANDOM_SEARCH'),
  ])
  def testCreationFromAndToProtoStudyStringAlgorithm(self, algorithm):
    study_config_proto = study_pb2.StudySpec(
        metrics=[
            study_pb2.StudySpec.MetricSpec(
                metric_id='pr-auc',
                goal=study_pb2.StudySpec.MetricSpec.GoalType.MAXIMIZE)
        ],
        algorithm=algorithm)
    # Test the algorithm when pythia endpoint is not specified.
    # This can be used when a pythia service is injected directly to the Vizier
    # service class.
    sc = vz.StudyConfig.from_proto(study_config_proto)
    self.assertEqual(sc.algorithm, algorithm)
    self.assertIsNone(sc.pythia_endpoint)

    compare.assertProto2Equal(self, study_config_proto, sc.to_proto())
    _ = vz.StudyConfig.from_problem(sc.to_problem())  # smoke test.

  @parameterized.parameters([
      'my custom algorithm',
      'QUASI_RANDOM_SEARCH',
  ])
  def testCreationFromAndToProtoStudyStringAlgorithmPythiaEndpoint(
      self, algorithm
  ):
    study_config_proto = study_pb2.StudySpec(
        metrics=[
            study_pb2.StudySpec.MetricSpec(
                metric_id='pr-auc',
                goal=study_pb2.StudySpec.MetricSpec.GoalType.MAXIMIZE,
            )
        ],
        algorithm=algorithm,
        metadata=[
            key_value_pb2.KeyValue(
                key=constants.PYTHIA_ENDPOINT_KEY,
                ns=vz.Namespace([constants.PYTHIA_ENDPOINT_NAMESPACE]).encode(),
                value='localhost:8888',
            )
        ],
    )
    sc = vz.StudyConfig.from_proto(study_config_proto)
    self.assertEqual(sc.algorithm, algorithm)
    self.assertEqual(sc.pythia_endpoint, 'localhost:8888')

    compare.assertProto2Equal(self, study_config_proto, sc.to_proto())
    _ = vz.StudyConfig.from_problem(sc.to_problem())  # smoke test.

  def testCreationFromAndToProtoMultiObjectiveStudy(self):
    study_config_proto = study_pb2.StudySpec(
        metrics=[
            study_pb2.StudySpec.MetricSpec(
                metric_id='pr-auc',
                goal=study_pb2.StudySpec.MetricSpec.GoalType.MAXIMIZE),
            study_pb2.StudySpec.MetricSpec(
                metric_id='loss',
                goal=study_pb2.StudySpec.MetricSpec.GoalType.MINIMIZE)
        ],)
    study_config_proto.parameters.extend(self.pconfigs)
    # Test all proprties.
    sc = vz.StudyConfig.from_proto(study_config_proto)

    expected = vz.MetricsConfig([
        vz.MetricInformation(name='loss', goal=vz.ObjectiveMetricGoal.MINIMIZE),
        vz.MetricInformation(
            name='pr-auc', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        ),
    ])
    self.assertEqual(sc.metric_information, expected)
    self.assertIsNone(sc.single_objective_metric_name)
    self.assertFalse(sc.is_single_objective)

    round_trip_proto = sc.to_proto()
    compare.assertProto2SameElements(self, study_config_proto, round_trip_proto)

    _ = vz.StudyConfig.from_problem(sc.to_problem())  # smoke test.

  def testCreationFromAndToProtoSafeStudy(self):
    expected_automated_stopping_config = (
        study_pb2.StudySpec.DefaultEarlyStoppingSpec())

    study_config_proto = study_pb2.StudySpec(
        algorithm='QUASI_RANDOM_SEARCH',
        metrics=[
            study_pb2.StudySpec.MetricSpec(
                metric_id='pr-auc',
                goal=study_pb2.StudySpec.MetricSpec.GoalType.MAXIMIZE),
            study_pb2.StudySpec.MetricSpec(
                metric_id='privacy-safety',
                goal=study_pb2.StudySpec.MetricSpec.GoalType.MINIMIZE,
                safety_config=study_pb2.StudySpec.MetricSpec.SafetyMetricConfig(
                    safety_threshold=0.2, desired_min_safe_trials_fraction=0.8))
        ],
        default_stopping_spec=expected_automated_stopping_config)

    study_config_proto.parameters.extend(self.pconfigs)
    # Test all proprties.
    sc = vz.StudyConfig.from_proto(study_config_proto)
    expected = vz.MetricsConfig([
        vz.MetricInformation(
            name='pr-auc', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        ),
        vz.MetricInformation(
            name='privacy-safety',
            goal=vz.ObjectiveMetricGoal.MINIMIZE,
            safety_threshold=0.2,
            desired_min_safe_trials_fraction=0.8,
        ),
    ])
    self.assertEqual(sc.metric_information, expected)
    self.assertEqual(sc.algorithm, 'QUASI_RANDOM_SEARCH')
    self.assertIsNone(sc.pythia_endpoint)
    self.assertEqual(sc.single_objective_metric_name, 'pr-auc')
    self.assertTrue(sc.is_single_objective)

    compare.assertProto2Equal(self, expected_automated_stopping_config,
                              sc.automated_stopping_config.to_proto())
    compare.assertProto2Equal(self, study_config_proto, sc.to_proto())
    _ = vz.StudyConfig.from_problem(sc.to_problem())  # smoke test.

  def testCreationFromProtoNoGoalRaises(self):
    study_config_proto = study_pb2.StudySpec()

    sc = vz.StudyConfig.from_proto(study_config_proto)
    self.assertEmpty(sc.metric_information)

  def testMetadata(self):
    empty_trial = study_pb2.Trial(id=str(1))
    sc = vz.StudyConfig()
    sc.metric_information.append(
        vz.MetricInformation(
            name='pr-auc', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )
    sc.metadata.abs_ns(['ns'])['key'] = 'ns-value'
    sc.metadata.abs_ns()['key'] = 'value'
    sc.metadata['proto'] = empty_trial

    proto = sc.to_proto()
    compare.assertProto2Contains(
        self, """metadata {
        key: "key"
        ns: ":ns"
        value: "ns-value"
      }
      metadata {
        key: "key"
        value: "value"
      }
      metadata {
        key: "proto"
        proto {
          [type.googleapis.com/vizier.Trial] {
            id: '1'
          }
        }
      }
    """, proto)
    from_proto = sc.from_proto(proto).metadata
    self.assertCountEqual(sc.metadata, from_proto)

  def testCreation(self):
    sc = vz.StudyConfig()
    sc.algorithm = vz.Algorithm.RANDOM_SEARCH
    sc.metric_information.append(
        vz.MetricInformation(
            name='pr-auc', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )
    root = sc.search_space.root
    root.add_float_param(
        'learning_rate', 0.00001, 1.0, scale_type=vz.ScaleType.LINEAR
    )
    root.add_categorical_param('optimizer', ['adagrad', 'adam', 'experimental'])

    sc.automated_stopping_config = (
        vz.AutomatedStoppingConfig.default_stopping_spec()
    )

    # Test all proprties.
    self.assertEqual(sc.algorithm, 'RANDOM_SEARCH')
    expected = vz.MetricsConfig(
        [
            vz.MetricInformation(
                name='pr-auc', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    self.assertEqual(sc.metric_information, expected)
    self.assertEqual(sc.single_objective_metric_name, 'pr-auc')
    self.assertTrue(sc.is_single_objective)

    expected = study_pb2.StudySpec(
        algorithm='RANDOM_SEARCH',
        metrics=[
            study_pb2.StudySpec.MetricSpec(
                metric_id='pr-auc',
                goal=study_pb2.StudySpec.MetricSpec.GoalType.MAXIMIZE)
        ],
        default_stopping_spec=study_pb2.StudySpec.DefaultEarlyStoppingSpec(),
        observation_noise=study_pb2.StudySpec.ObservationNoise
        .OBSERVATION_NOISE_UNSPECIFIED,
    )
    expected.parameters.extend(self.pconfigs)
    compare.assertProto2Equal(self, expected, sc.to_proto())

  @absltest.skip('???')
  def testTrialToDict(self):
    py_study_config = vz.StudyConfig(
        metric_information=[
            vz.MetricInformation(
                name='objective', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    root = py_study_config.search_space.root
    root.add_float_param('learning_rate', 0.01, 3.0)
    root.add_int_param('units', 10, 1000, scale_type=vz.ScaleType.LOG)
    root.add_discrete_param('batch_size', [8, 16, 32])
    root.add_discrete_param(
        'floating_point_param', [8., 16., 32.], auto_cast=False)
    root.add_categorical_param('activation', ['tanh', 'relu'])
    root.add_bool_param('synchronous')

    trial_proto = study_pb2.Trial(id=str(1))
    trial_proto.parameters.add(
        parameter_id='activation', value=struct_pb2.Value(string_value='relu'))
    trial_proto.parameters.add(
        parameter_id='synchronus', value=struct_pb2.Value(string_value='true'))
    trial_proto.parameters.add(
        parameter_id='batch_size', value=struct_pb2.Value(number_value=32))
    trial_proto.parameters.add(
        parameter_id='floating_point_param',
        value=struct_pb2.Value(number_value=32))
    trial_proto.parameters.add(
        parameter_id='learning_rate', value=struct_pb2.Value(number_value=0.5))
    trial_proto.parameters.add(
        parameter_id='units', value=struct_pb2.Value(number_value=50))

    parameters = py_study_config.trial_parameters(trial_proto)
    expected = {
        'learning_rate': 0.5,
        'units': 50,
        'activation': 'relu',
        'batch_size': 32,
        'floating_point_param': 32.,
        'synchronous': True
    }
    self.assertEqual(expected, parameters)
    self.assertIsInstance(parameters['batch_size'], int)
    self.assertIsInstance(parameters['floating_point_param'], float)

  def testPyTrialToDict(self):
    py_study_config = vz.StudyConfig(
        metric_information=[
            vz.MetricInformation(
                name='objective', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    root = py_study_config.search_space.root
    root.add_float_param('learning_rate', 0.01, 3.0)
    root.add_int_param('units', 10, 1000, scale_type=vz.ScaleType.LOG)
    root.add_discrete_param('batch_size', [8, 16, 32])
    root.add_discrete_param(
        'floating_point_param', [8., 16., 32.], auto_cast=False)
    root.add_categorical_param('activation', ['tanh', 'relu'])
    root.add_bool_param('synchronous')

    pytrial = vz.Trial(id=1)
    pytrial.parameters = {
        'activation': vz.ParameterValue(value='relu'),
        'synchronous': vz.ParameterValue(value=True),
        'batch_size': vz.ParameterValue(value=32),
        'floating_point_param': vz.ParameterValue(value=32.0),
        'learning_rate': vz.ParameterValue(value=0.5),
        'units': vz.ParameterValue(value=50),
    }
    parameters = py_study_config._pytrial_parameters(pytrial)
    expected = {
        'learning_rate': 0.5,
        'units': 50,
        'activation': 'relu',
        'batch_size': 32,
        'floating_point_param': 32.,
        'synchronous': True
    }
    self.assertEqual(expected, parameters)
    self.assertIsInstance(parameters['batch_size'], int)
    self.assertIsInstance(parameters['floating_point_param'], float)

  def testTrialToDictWithoutExternalType(self):
    """Test conversion when external types are not specified."""
    proto = study_pb2.StudySpec()
    proto.parameters.add(
        parameter_id='learning_rate',
        double_value_spec=study_pb2.StudySpec.ParameterSpec.DoubleValueSpec(
            min_value=1e-4, max_value=0.1),
        scale_type=study_pb2.StudySpec.ParameterSpec.ScaleType.UNIT_LOG_SCALE)
    proto.parameters.add(
        parameter_id='batch_size',
        discrete_value_spec=study_pb2.StudySpec.ParameterSpec.DiscreteValueSpec(
            values=[1.0, 2.0, 4.0, 8.0, 16.0, 32.0, 64.0, 128.0]))
    proto.parameters.add(
        parameter_id='training_steps',
        discrete_value_spec=study_pb2.StudySpec.ParameterSpec.DiscreteValueSpec(
            values=[1000.0, 10000.0]))
    proto.observation_noise = study_pb2.StudySpec.ObservationNoise.HIGH
    proto.metrics.add(
        metric_id='loss', goal=study_pb2.StudySpec.MetricSpec.MINIMIZE)

    trial_proto = study_pb2.Trial()
    trial_proto.id = str(1)
    trial_proto.parameters.add(
        parameter_id='batch_size', value=struct_pb2.Value(number_value=128.0))
    trial_proto.parameters.add(
        parameter_id='learning_rate',
        value=struct_pb2.Value(number_value=1.2137854406366652E-4))
    trial_proto.parameters.add(
        parameter_id='training_steps',
        value=struct_pb2.Value(number_value=10000.0))

    py_study_config = vz.StudyConfig.from_proto(proto)
    self.assertEqual(
        py_study_config.observation_noise, vz.ObservationNoise.HIGH
    )
    parameters = py_study_config.trial_parameters(trial_proto)
    self.assertEqual(
        py_study_config.observation_noise, vz.ObservationNoise.HIGH
    )
    expected = {
        'batch_size': 128,
        'learning_rate': 1.2137854406366652E-4,
        'training_steps': 10000.0
    }
    self.assertEqual(expected, parameters)
    self.assertIsInstance(parameters['learning_rate'], float)
    self.assertIsInstance(parameters['batch_size'], float)
    self.assertIsInstance(parameters['training_steps'], float)

  @absltest.skip('???')
  def testTrialToDictMultidimensional(self):
    py_study_config = vz.StudyConfig(
        metric_information=[
            vz.MetricInformation(
                name='objective', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    root = py_study_config.search_space.root
    for index in (0, 1):
      root.add_float_param('learning_rate', 0.01, 3.0, index=index)
      root.add_int_param(
          'units', 10, 1000, scale_type=vz.ScaleType.LOG, index=index
      )
      root.add_categorical_param('activation', ['tanh', 'relu'], index=index)
      root.add_bool_param('synchronous', index=index)
      root.add_discrete_param('batch_size', [8, 16, 32], index=index)
    root.add_discrete_param(
        'floating_point_param', [8., 16., 32.], auto_cast=False)

    trial_proto = study_pb2.Trial()
    trial_proto.id = str(2)
    trial_proto.parameters.add(
        parameter_id='learning_rate[0]',
        value=struct_pb2.Value(number_value=0.5))
    trial_proto.parameters.add(
        parameter_id='learning_rate[1]',
        value=struct_pb2.Value(number_value=0.1))
    trial_proto.parameters.add(
        parameter_id='units[0]', value=struct_pb2.Value(number_value=50))
    trial_proto.parameters.add(
        parameter_id='units[1]', value=struct_pb2.Value(number_value=200))
    trial_proto.parameters.add(
        parameter_id='activation[0]',
        value=struct_pb2.Value(string_value='relu'))
    trial_proto.parameters.add(
        parameter_id='activation[1]',
        value=struct_pb2.Value(string_value='relu'))
    trial_proto.parameters.add(
        parameter_id='synchronus[0]',
        value=struct_pb2.Value(string_value='true'))
    trial_proto.parameters.add(
        parameter_id='synchronus[1]',
        value=struct_pb2.Value(string_value='false'))
    trial_proto.parameters.add(
        parameter_id='batch_size[0]', value=struct_pb2.Value(number_value=32.0))
    trial_proto.parameters.add(
        parameter_id='batch_size[1]', value=struct_pb2.Value(number_value=8.0))
    trial_proto.parameters.add(
        parameter_id='floating_point_param',
        value=struct_pb2.Value(number_value=16.0))
    parameters = py_study_config.trial_parameters(trial_proto)
    expected = {
        'learning_rate': [0.5, 0.1],
        'units': [50, 200],
        'activation': ['relu', 'relu'],
        'batch_size': [32, 8],
        'synchronous': [True, False],
        'floating_point_param': 16.,
    }
    self.assertEqual(expected, parameters)

  def testPyTrialToDictMultidimensional(self):
    py_study_config = vz.StudyConfig(
        metric_information=[
            vz.MetricInformation(
                name='objective', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    root = py_study_config.search_space.root
    for index in (0, 1):
      root.add_float_param('learning_rate', 0.01, 3.0, index=index)
      root.add_int_param(
          'units', 10, 1000, scale_type=vz.ScaleType.LOG, index=index
      )
      root.add_categorical_param('activation', ['tanh', 'relu'], index=index)
      root.add_bool_param('synchronous', index=index)
      root.add_discrete_param('batch_size', [8, 16, 32], index=index)
    root.add_discrete_param(
        'floating_point_param', [8., 16., 32.], auto_cast=False)

    pytrial = vz.Trial(id=2)
    pytrial.parameters = {
        'learning_rate[0]': vz.ParameterValue(value=0.5),
        'learning_rate[1]': vz.ParameterValue(value=0.1),
        'units[0]': vz.ParameterValue(value=50),
        'units[1]': vz.ParameterValue(value=200),
        'activation[0]': vz.ParameterValue(value='relu'),
        'activation[1]': vz.ParameterValue(value='relu'),
        'synchronous[0]': vz.ParameterValue(value=True),
        'synchronous[1]': vz.ParameterValue(value=False),
        'batch_size[0]': vz.ParameterValue(value=32.0),
        'batch_size[1]': vz.ParameterValue(value=8.0),
        'floating_point_param': vz.ParameterValue(value=16.0),
    }
    parameters = py_study_config._pytrial_parameters(pytrial)
    expected = {
        'learning_rate': [0.5, 0.1],
        'units': [50, 200],
        'activation': ['relu', 'relu'],
        'batch_size': [32, 8],
        'synchronous': [True, False],
        'floating_point_param': 16.,
    }
    self.assertEqual(expected, parameters)

  def testGinConfigMultiDimensional(self):
    py_study_config = vz.StudyConfig(
        metric_information=[
            vz.MetricInformation(
                name='objective', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    root = py_study_config.search_space.root
    block_categories = [
        'block_3x3', 'block_4x4', 'block_1x3_3x1', 'block_1x3_3x1_dw',
        'block_identity'
    ]
    for index in range(5):
      root.add_categorical_param(
          '_gin.ambient_net_exp_from_vec.block_type',
          feasible_values=block_categories,
          index=index)

    trial_proto = study_pb2.Trial()
    trial_proto.id = str(2)
    trial_proto.parameters.add(
        parameter_id='_gin.ambient_net_exp_from_vec.block_type[0]',
        value=struct_pb2.Value(string_value='block_3x3'))
    trial_proto.parameters.add(
        parameter_id='_gin.ambient_net_exp_from_vec.block_type[1]',
        value=struct_pb2.Value(string_value='block_4x4'))
    trial_proto.parameters.add(
        parameter_id='_gin.ambient_net_exp_from_vec.block_type[2]',
        value=struct_pb2.Value(string_value='block_1x3_3x1_dw'))
    trial_proto.parameters.add(
        parameter_id='_gin.ambient_net_exp_from_vec.block_type[3]',
        value=struct_pb2.Value(string_value='block_identity'))
    trial_proto.parameters.add(
        parameter_id='_gin.ambient_net_exp_from_vec.block_type[4]',
        value=struct_pb2.Value(string_value='block_1x3_3x1'))

    parameters = py_study_config.trial_parameters(trial_proto)
    expected = {
        '_gin.ambient_net_exp_from_vec.block_type': [
            'block_3x3', 'block_4x4', 'block_1x3_3x1_dw', 'block_identity',
            'block_1x3_3x1'
        ],
    }
    self.assertEqual(expected, parameters)

  @absltest.skip('???')
  def testTrialToDictConditional(self):
    py_study_config = vz.StudyConfig(
        metric_information=[
            vz.MetricInformation(
                name='objective', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    root = py_study_config.search_space.root

    model_type = root.add_categorical_param('model_type', ['dnn', 'linear'])
    dnn = model_type.select_values(['dnn'])
    dnn.add_float_param('learning_rate', 0.01, 3.0)
    dnn.add_int_param('units', 1, 50, index=0)
    dnn.add_int_param('units', 1, 80, index=1)
    dnn.add_categorical_param('activation', ['tanh', 'relu'])
    model_type.select_values(['linear'
                             ]).add_float_param('learning_rate', 0.01, 1.0)

    dnn_trial = study_pb2.Trial()
    dnn_trial.id = str(1)
    dnn_trial.parameters.add(
        parameter_id='model_type', value=struct_pb2.Value(string_value='dnn'))
    dnn_trial.parameters.add(
        parameter_id='learning_rate', value=struct_pb2.Value(number_value=2.1))
    dnn_trial.parameters.add(
        parameter_id='unts[0]', value=struct_pb2.Value(number_value=49))
    dnn_trial.parameters.add(
        parameter_id='unts[1]', value=struct_pb2.Value(number_value=79))
    dnn_trial.parameters.add(
        parameter_id='activation', value=struct_pb2.Value(string_value='relu'))

    parameters = py_study_config.trial_parameters(dnn_trial)
    expected = {
        'model_type': 'dnn',
        'learning_rate': 2.1,
        'units': [49, 79],
        'activation': 'relu',
    }
    self.assertEqual(expected, parameters)

  def testPyTrialToDictConditional(self):
    py_study_config = vz.StudyConfig(
        metric_information=[
            vz.MetricInformation(
                name='objective', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    root = py_study_config.search_space.root

    model_type = root.add_categorical_param('model_type', ['dnn', 'linear'])
    dnn = model_type.select_values(['dnn'])
    dnn.add_float_param('learning_rate', 0.01, 3.0)
    dnn.add_int_param('units', 1, 50, index=0)
    dnn.add_int_param('units', 1, 80, index=1)
    dnn.add_categorical_param('activation', ['tanh', 'relu'])
    model_type.select_values(['linear'
                             ]).add_float_param('learning_rate', 0.01, 1.0)

    pytrial = vz.Trial(
        id=1,
        parameters={
            'model_type': vz.ParameterValue(value='dnn'),
            'learning_rate': vz.ParameterValue(value=2.1),
            'units[0]': vz.ParameterValue(value=49),
            'units[1]': vz.ParameterValue(value=79),
            'activation': vz.ParameterValue(value='relu'),
        },
    )
    parameters = py_study_config._pytrial_parameters(pytrial)
    expected = {
        'model_type': 'dnn',
        'learning_rate': 2.1,
        'units': [49, 79],
        'activation': 'relu',
    }
    self.assertEqual(expected, parameters)

  def testTrialToDictRaisesDuplicateParameters(self):
    py_study_config = vz.StudyConfig(
        metric_information=[
            vz.MetricInformation(
                name='objective', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    trial_proto = study_pb2.Trial()
    trial_proto.id = str(1)
    trial_proto.parameters.add(
        parameter_id='activation', value=struct_pb2.Value(string_value='relu'))
    trial_proto.parameters.add(
        parameter_id='activation', value=struct_pb2.Value(string_value='tanh'))
    trial_proto.parameters.add(
        parameter_id='units', value=struct_pb2.Value(number_value=50))

    with self.assertRaisesRegex(ValueError, 'Invalid trial proto'):
      py_study_config.trial_parameters(trial_proto)

  def testTrialToDictRaisesInvalidTrial(self):
    py_study_config = vz.StudyConfig(
        metric_information=[
            vz.MetricInformation(
                name='objective', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    root = py_study_config.search_space.root
    root.add_float_param('learning_rate', 0.01, 3.0)

    trial_proto = study_pb2.Trial()
    trial_proto.id = str(1)
    trial_proto.state = study_pb2.Trial.State.ACTIVE
    trial_proto.parameters.add(
        parameter_id='foo', value=struct_pb2.Value(number_value=0.5))
    with self.assertRaisesRegex(ValueError,
                                'Invalid trial for this search space'):
      py_study_config.trial_parameters(trial_proto)

  def testTrialToDictWithFinalMetricsSingleObjective(self):
    py_study_config = vz.StudyConfig(
        metric_information=[
            vz.MetricInformation(
                name='objective', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    root = py_study_config.search_space.root
    root.add_float_param('learning_rate', 0.01, 3.0)

    trial_proto = study_pb2.Trial()
    trial_proto.id = str(1)
    trial_proto.state = study_pb2.Trial.State.SUCCEEDED
    trial_proto.parameters.add(
        parameter_id='learning_rate', value=struct_pb2.Value(number_value=0.5))
    trial_proto.final_measurement.step_count = 101
    trial_proto.final_measurement.elapsed_duration.seconds = 67
    trial_proto.final_measurement.metrics.add(metric_id='loss', value=56.8)
    trial_proto.final_measurement.metrics.add(metric_id='objective', value=77.7)

    parameters = py_study_config.trial_parameters(trial_proto)
    self.assertEqual({'learning_rate': 0.5}, parameters)
    metrics = py_study_config.trial_metrics(trial_proto)
    self.assertEqual({'objective': 77.7}, metrics)
    metrics = py_study_config.trial_metrics(
        trial_proto, include_all_metrics=True)
    self.assertEqual({'objective': 77.7, 'loss': 56.8}, metrics)

  def testPyTrialToDictWithFinalMetricsSingleObjective(self):
    py_study_config = vz.StudyConfig(
        metric_information=[
            vz.MetricInformation(
                name='objective', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    root = py_study_config.search_space.root
    root.add_float_param('learning_rate', 0.01, 3.0)

    pytrial = vz.Trial(
        id=1,
        completion_time=datetime.datetime(
            year=2021, month=12, day=2, hour=7, minute=31
        ),
        parameters={'learning_rate': vz.ParameterValue(0.5)},
        final_measurement=vz.Measurement(
            metrics={
                'loss': vz.Metric(value=56.8),
                'objective': vz.Metric(value=77.7),
            },
            elapsed_secs=67,
            steps=101,
        ),
    )
    parameters = py_study_config._pytrial_parameters(pytrial)
    self.assertEqual({'learning_rate': 0.5}, parameters)
    metrics = py_study_config._pytrial_metrics(pytrial)
    self.assertEqual({'objective': 77.7}, metrics)
    metrics = py_study_config._pytrial_metrics(
        pytrial, include_all_metrics=True)
    self.assertEqual({'objective': 77.7, 'loss': 56.8}, metrics)

  def testTrialToDictWithFinalMetricsNotCompleted(self):
    # Throw a Trial that has inconsistent field values.
    # (ACTIVE but has final measurement).
    # Pyvizier fixes the state.
    py_study_config = vz.StudyConfig(
        metric_information=[
            vz.MetricInformation(
                name='objective', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    root = py_study_config.search_space.root
    root.add_float_param('learning_rate', 0.01, 3.0)

    trial_proto = study_pb2.Trial()
    trial_proto.id = str(1)
    trial_proto.state = study_pb2.Trial.State.ACTIVE
    trial_proto.parameters.add(
        parameter_id='learning_rate', value=struct_pb2.Value(number_value=0.5))
    trial_proto.final_measurement.step_count = 101
    trial_proto.final_measurement.elapsed_duration.seconds = 67
    trial_proto.final_measurement.metrics.add(metric_id='loss', value=56.8)
    trial_proto.final_measurement.metrics.add(metric_id='objective', value=77.7)

    parameters = py_study_config.trial_parameters(trial_proto)
    self.assertEqual({'learning_rate': 0.5}, parameters)
    self.assertLen(
        py_study_config.trial_metrics(trial_proto, include_all_metrics=True), 2)

  def testTrialToDictWithFinalMetricsInfeasible(self):
    py_study_config = vz.StudyConfig(
        metric_information=[
            vz.MetricInformation(
                name='objective', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    root = py_study_config.search_space.root
    root.add_float_param('learning_rate', 0.01, 3.0)

    trial_proto = study_pb2.Trial()
    trial_proto.id = str(1)
    trial_proto.state = study_pb2.Trial.State.INFEASIBLE
    trial_proto.parameters.add(
        parameter_id='learning_rate', value=struct_pb2.Value(number_value=0.5))
    trial_proto.final_measurement.step_count = 101
    trial_proto.final_measurement.elapsed_duration.seconds = 67
    trial_proto.final_measurement.metrics.add(metric_id='loss', value=56.8)
    trial_proto.final_measurement.metrics.add(metric_id='objective', value=77.7)

    parameters = py_study_config.trial_parameters(trial_proto)
    self.assertEqual({'learning_rate': 0.5}, parameters)
    self.assertEmpty(py_study_config.trial_metrics(trial_proto))
    self.assertEmpty(
        py_study_config.trial_metrics(trial_proto, include_all_metrics=True))

  def testPyTrialToDictWithFinalMetricsInfeasible(self):
    py_study_config = vz.StudyConfig(
        metric_information=[
            vz.MetricInformation(
                name='objective', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    root = py_study_config.search_space.root
    root.add_float_param('learning_rate', 0.01, 3.0)

    pytrial = vz.Trial(
        id=1,
        infeasibility_reason='just because',
        completion_time=datetime.datetime(
            year=2021, month=12, day=2, hour=7, minute=31
        ),
        parameters={'learning_rate': vz.ParameterValue(0.5)},
        final_measurement=vz.Measurement(
            metrics={
                'loss': vz.Metric(value=56.8),
                'other': vz.Metric(value=77.7),
            },
            elapsed_secs=67,
            steps=101,
        ),
    )
    parameters = py_study_config._pytrial_parameters(pytrial)
    self.assertEqual({'learning_rate': 0.5}, parameters)
    self.assertEmpty(py_study_config._pytrial_metrics(pytrial))
    self.assertEmpty(
        py_study_config._pytrial_metrics(pytrial, include_all_metrics=True))

  def testTrialToDictWithFinalMetricsMultiObjective(self):
    py_study_config = vz.StudyConfig(
        metric_information=[
            vz.MetricInformation(
                name='objective', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            ),
            vz.MetricInformation(
                name='objective2', goal=vz.ObjectiveMetricGoal.MINIMIZE
            ),
        ]
    )
    root = py_study_config.search_space.root
    root.add_float_param('learning_rate', 0.01, 3.0)

    trial_proto = study_pb2.Trial(id=str(1))
    trial_proto.state = study_pb2.Trial.State.SUCCEEDED

    trial_proto.parameters.add(
        parameter_id='learning_rate', value=struct_pb2.Value(number_value=0.5))
    trial_proto.final_measurement.step_count = 101
    trial_proto.final_measurement.elapsed_duration.seconds = 67
    trial_proto.final_measurement.metrics.add(metric_id='loss', value=56.8)
    trial_proto.final_measurement.metrics.add(metric_id='objective', value=77.7)
    trial_proto.final_measurement.metrics.add(
        metric_id='objective2', value=-0.2)

    parameters = py_study_config.trial_parameters(trial_proto)
    self.assertEqual({'learning_rate': 0.5}, parameters)
    metrics = py_study_config.trial_metrics(trial_proto)
    self.assertEqual({'objective': 77.7, 'objective2': -0.2}, metrics)
    metrics = py_study_config.trial_metrics(
        trial_proto, include_all_metrics=True)
    self.assertEqual({
        'objective': 77.7,
        'objective2': -0.2,
        'loss': 56.8
    }, metrics)

  def testPyTrialToDictWithFinalMetricsMultiObjective(self):
    py_study_config = vz.StudyConfig(
        metric_information=[
            vz.MetricInformation(
                name='objective', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            ),
            vz.MetricInformation(
                name='objective2', goal=vz.ObjectiveMetricGoal.MINIMIZE
            ),
        ]
    )
    root = py_study_config.search_space.root
    root.add_float_param('learning_rate', 0.01, 3.0)

    pytrial = vz.Trial(
        id=1,
        completion_time=datetime.datetime(
            year=2021, month=12, day=2, hour=7, minute=31
        ),
        parameters={'learning_rate': vz.ParameterValue(0.5)},
        final_measurement=vz.Measurement(
            metrics={
                'loss': vz.Metric(value=56.8),
                'objective': vz.Metric(value=77.7),
                'objective2': vz.Metric(value=-0.2),
            },
            elapsed_secs=67,
            steps=101,
        ),
    )
    parameters = py_study_config._pytrial_parameters(pytrial)
    self.assertEqual({'learning_rate': 0.5}, parameters)
    metrics = py_study_config._pytrial_metrics(pytrial)
    self.assertEqual({'objective': 77.7, 'objective2': -0.2}, metrics)
    metrics = py_study_config._pytrial_metrics(
        pytrial, include_all_metrics=True)
    self.assertEqual({
        'objective': 77.7,
        'objective2': -0.2,
        'loss': 56.8
    }, metrics)

  def testSearchSpacesNotShared(self):
    sc1 = vz.StudyConfig()
    sc1.search_space.root.add_float_param('x', 1, 2)
    sc2 = vz.StudyConfig()
    sc2.search_space.root.add_float_param('x', 1, 2)
    self.assertLen(sc1.search_space.parameters, 1)
    self.assertLen(sc2.search_space.parameters, 1)

  def testHasConditionalParametersFlatSpace(self):
    sc = vz.StudyConfig()
    sc.search_space.root.add_float_param('x', 1, 2)
    self.assertFalse(sc.search_space.is_conditional)

  def testHasConditionalParameters(self):
    sc = vz.StudyConfig()
    root = sc.search_space.root
    model_type = root.add_categorical_param('model_type', ['linear', 'dnn'])
    _ = model_type.select_values(['dnn']).add_float_param(
        'learning_rate',
        0.1,
        1.0,
        default_value=0.001,
        scale_type=vz.ScaleType.LOG,
    )
    self.assertTrue(sc.search_space.is_conditional)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for combo_experimenter."""
# pylint:disable=g-long-lambda
import functools
from absl import logging

from vizier import pyvizier
from vizier._src.algorithms.designers import random
from vizier._src.benchmarks.experimenters import combo_experimenter

from absl.testing import absltest
from absl.testing import parameterized


class ComboExperimenterTest(parameterized.TestCase):
  """Default lamda values found in README of https://github.com/QUVA-Lab/COMBO."""

  @parameterized.named_parameters(
      ('contamination',
       functools.partial(
           combo_experimenter.ContaminationExperimenter,
           lamda=0.01), 20.0, 30.0),
      ('ising',
       functools.partial(combo_experimenter.IsingExperimenter,
                         lamda=0.01), 0.0, 50.0),
      ('centroid', combo_experimenter.CentroidExperimenter, 0.0, 150.0),
      ('pest_control', combo_experimenter.PestControlExperimenter, 12.0, 25.0),
  )
  def test_experimenters(self, experimenter_class, objective_min,
                         objective_max):
    """Tests if entire pipeline works and objective values make sense."""
    experimenter = experimenter_class()

    problem_statement = experimenter.problem_statement()
    logging.info(problem_statement)
    designer = random.RandomDesigner(
        search_space=problem_statement.search_space, seed=None)

    suggestions = designer.suggest(5)
    trials = [suggestion.to_trial() for suggestion in suggestions]
    experimenter.evaluate(trials)
    for trial in trials:
      logging.info('Evaluated Trial: %s', trial)
      self.assertEqual(trial.status, pyvizier.TrialStatus.COMPLETED)
      metric_name = problem_statement.metric_information.item().name
      eval_objective = trial.final_measurement.metrics[metric_name].value
      self.assertLessEqual(eval_objective, objective_max)
      self.assertGreaterEqual(eval_objective, objective_min)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""SimpleKD convergence tests."""

from absl import logging
import attrs
from vizier import algorithms as vza
from vizier._src.benchmarks.experimenters.synthetic import simplekd
from vizier._src.benchmarks.runners import benchmark_runner
from vizier._src.benchmarks.runners import benchmark_state


class FailedSimpleKDConvergenceTestError(Exception):
  """Exception raised for simplekd convergence test fails."""


@attrs.define(kw_only=True)
class SimpleKDConvergenceTester:
  """The class tests designers convergence on SimpleKD problem."""

  # The SimpleKD problem type.
  best_category: simplekd.SimpleKDCategory

  # A designer factory to generate different instances of the designer.
  designer_factory: vza.DesignerFactory[vza.Designer]

  # The number of the suggested trials in each run.
  num_trials: int

  # The maximum objective value relative error to determine convergence.
  max_relative_error: float

  # The number of repeated runs used to obtain reliable results.
  num_repeats: int

  # The target number of converged runs.
  target_num_convergence: int

  # The number of designer trials suggested in each repeat.
  batch_size: int = 1

  # The number of float parameters.
  num_float_param: int = 1

  # The number of discrete parameters.
  num_discrete_param: int = 1

  # The number of integer parameters.
  num_int_param: int = 1

  # Whether to use random seeds when instantiating the designer.
  is_deterministic: bool = True

  def assert_convergence(self) -> None:
    """Run the convergence test."""
    exptr = simplekd.SimpleKD(
        self.best_category,
        num_float_param=self.num_float_param,
        num_discrete_param=self.num_discrete_param,
        num_int_param=self.num_int_param,
        output_relative_error=True,
    )
    runner = benchmark_runner.BenchmarkRunner(
        benchmark_subroutines=[
            benchmark_runner.GenerateAndEvaluate(batch_size=self.batch_size),
        ],
        num_repeats=self.num_trials // self.batch_size,
    )
    num_success = 0
    for seed in range(self.num_repeats):
      state = benchmark_state.BenchmarkState(
          exptr,
          benchmark_state.PolicySuggester.from_designer_factory(
              problem=exptr.problem_statement(),
              designer_factory=self.designer_factory,
              seed=seed if self.is_deterministic else None,
          ),
      )
      runner.run(state)
      best_trial = state.algorithm.supporter.GetBestTrials(count=1)[0]
      metric_name = exptr.problem_statement().metric_information.item().name
      best_err = best_trial.final_measurement_or_die.metrics[metric_name].value
      if best_err <= self.max_relative_error:
        num_success += 1
    if num_success < self.target_num_convergence:
      raise FailedSimpleKDConvergenceTestError(
          'The simplekd convergence test failed. The number of converged runs'
          f' is {num_success}, though expected at least'
          f' {self.target_num_convergence} converged runs.'
      )
    else:
      logging.info(
          'The simplekd convergence test passed. The number of converged runs'
          ' is %s, and the expected number of converged runs is %s.',
          num_success,
          self.target_num_convergence,
      )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""HPO-B dataset.

Note that we denote (X,Y) as a batched set of trials (i.e. suggestions X and
objectives Y) and (x,y) as a single trial. This is slightly different from (X,y)
notation used in the handler to denote batched trials.
"""
# TODO: Replace internal HPOB experimenter with this.
# pylint:disable=invalid-name
import copy
import enum
import json
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple

import attr
import attrs
import numpy as np

from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters import experimenter
from vizier._src.benchmarks.experimenters.hpob import handler as handler_lib

import xgboost as xgb

Open = open

METRIC_NAME = 'objective_value'
# Offset applied to parameter values before the log transformation.
_TF_OFFSET = 1e-4


@attrs.define(auto_attribs=True)
class _Dataset:
  """Raw data from HPO-B.

  X and Y are guaranteed to have compatible shapes. A dataset can be sliced like
  regular numpy arrays but but cannot be indexed at a single point, i.e.
    `dataset[0]` is not allowed
    `dataset[0:1]` is allowed
    `dataset[dataset.Y > 0]` is allowed.

  If the log-transformation is applied to a feature, it's offset by a constant
    x_log = np.log(x+0.0001)

  Attributes:
    X: 2-D array of shape (number of observations) * (number of input features).
      The features may be scaled and log-transformed. _SearchspaceDescriptor
      holds the necessary information to recover the original values.
    Y: 2-D array of objective values, of shape (number of observations, 1). The
      values are not pre-processed.
  """

  X: np.ndarray = attrs.field(converter=np.asarray)
  Y: np.ndarray = attrs.field(converter=np.asarray)

  def __attrs_post_init__(self) -> None:
    """Performs validation."""
    if len(self.X.shape) != 2:
      raise ValueError(f'X must be 2-D. Given: {self.X.shape}')
    if len(self.Y.shape) != 2:
      raise ValueError(f'Y must be 2-D. Given: {self.Y.shape}')

    if self.X.shape[0] != self.Y.shape[0]:
      raise ValueError(f'X and y must have same number of rows. '
                       f'X.shape={self.X.shape}, y.shape={self.Y.shape}')

  def __getitem__(self, idx: slice) -> '_Dataset':
    return _Dataset(self.X[idx], self.Y[idx])

  def __len__(self) -> int:
    return self.Y.shape[0]


@attr.define(init=True, kw_only=False)
class _VariableDescriptor:
  """Variable descriptor."""
  name: str = attrs.field()
  min_value: Optional[float] = attrs.field(default=None)
  max_value: Optional[float] = attrs.field(default=None)
  min_value_before_tf: Optional[float] = attrs.field(default=None)
  max_value_before_tf: Optional[float] = attrs.field(default=None)
  apply_log: bool = attrs.field(default=True)
  optional: bool = attrs.field(default=False)
  categories: List[str] = attrs.field(factory=list)

  def scale(self, x: np.ndarray) -> np.ndarray:
    # Raw numbers to HPOB scale. Input and output are 1-D
    if self.is_categorical:
      return x
    if self.apply_log:
      x = np.log(x)
    x = (x - self.min_value) / (self.max_value - self.min_value)
    return x

  def unscale(self, x: np.ndarray) -> np.ndarray:
    # HPOB scale to raw numbers. Input and output are 1-D.
    if self.is_categorical:
      return x
    x = x * (self.max_value - self.min_value) + self.min_value
    if self.apply_log:
      x = np.exp(x)
    bounds = self.bounds
    return np.clip(x, bounds[0], bounds[1])

  @classmethod
  def from_info(cls, name: str, info: Dict[str, Any], apply_log: bool,
                optional: bool) -> '_VariableDescriptor':
    if 'categories' in info:
      return cls(name, categories=info['categories'], optional=optional)
    else:
      return cls(
          name,
          min_value=info['min'],
          max_value=info['max'],
          min_value_before_tf=info['min_before_tf'],
          max_value_before_tf=info['max_before_tf'],
          apply_log=apply_log,
          optional=optional)

  @property
  def is_categorical(self) -> bool:
    return bool(self.categories)

  @property
  def bounds(self) -> Optional[Tuple[float, float]]:
    if self.is_categorical:
      bounds = None
    elif self.apply_log:
      bounds = (self.min_value_before_tf + _TF_OFFSET,
                self.max_value_before_tf + _TF_OFFSET)
    else:
      bounds = (self.min_value, self.max_value)
    return bounds  # pytype: disable=bad-return-type

  def as_parameter_config(self) -> vz.ParameterConfig:
    """Return the Vizier parameter config."""
    if self.is_categorical:
      return vz.ParameterConfig.factory(
          self.name, feasible_values=self.categories)
    else:
      if self.apply_log:
        scale_type = vz.ScaleType.LOG
      else:
        scale_type = vz.ScaleType.LINEAR
      return vz.ParameterConfig.factory(
          self.name, bounds=self.bounds, scale_type=scale_type)

  def as_discrete_parameter_config(self, feasible_values) -> vz.ParameterConfig:
    if self.is_categorical:
      raise ValueError('Only continuous parameters can be discretized.')
    else:
      pc = self.as_parameter_config()
      return vz.ParameterConfig.factory(
          pc.name, feasible_values=feasible_values, scale_type=pc.scale_type)


@attrs.define
class _SearchspaceDescriptor:
  variables: Dict[str, _VariableDescriptor]
  order: List[str]

  def column_index(self, name: str) -> int:
    return self.order.index(name)


class NaPolicy(enum.Enum):
  """Decides what to do given an optional parameter that may be missing.

    'drop': *_na does not exist in search space, but trials may be missing the
      associated parameter.
    'discrete': *_na is a separate parameter that takes value 0 or 1.
    'continuous': *_na is a separate parameter that takes value in [0, 1].
  """
  DROP = 'DROP'
  DISCRETE = 'DISCRETE'
  CONTINUOUS = 'CONTINUOUS'


class CategoricalPolicy(enum.Enum):
  """Decides what to do given a C-way categorical parameter.

    'as_categorical': Treat it as a categorical parameter.
    'as_continuous': Treat it as C continuous parameters in [0, 1], i.e.
      [0, 1]^C.
  """
  AS_CATEGORICAL = 'AS_CATEGORICAL'
  AS_CONTINUOUS = 'AS_CONTINUOUS'


class _HPOBVizierConverter:
  """Converts between HPOB and Vizier representations."""

  def __init__(
      self,
      descriptor: _SearchspaceDescriptor,
      data: Optional[_Dataset] = None,
      *,
      na_policy: NaPolicy = NaPolicy.DROP,
      categorical_policy: CategoricalPolicy = CategoricalPolicy.AS_CATEGORICAL):
    """Init.

    Args:
      descriptor: Search space descriptor.
      data: If not None, automatically identify discrete features.
      na_policy: See NaPolicy.
      categorical_policy: See CategoricalPolicy.

    Raises:
      ValueError:
    """
    self._descriptor = descriptor

    self.problem = vz.ProblemStatement()
    self.problem.metric_information.append(
        vz.MetricInformation(
            name=METRIC_NAME, goal=vz.ObjectiveMetricGoal.MAXIMIZE))
    self._na_policy = na_policy
    self._categorical_policy = categorical_policy

    if data is not None and data.X.shape[1] != len(descriptor.order):
      raise ValueError(
          f'data shape {data.X.shape} is not compatible with descriptor, which'
          f'defines {len(descriptor.order)} columns!')

    space = self.problem.search_space
    for variable in descriptor.variables.values():
      if variable.is_categorical:
        if self._categorical_policy == CategoricalPolicy.AS_CATEGORICAL:
          space.add(variable.as_parameter_config())
        elif self._categorical_policy == CategoricalPolicy.AS_CONTINUOUS:
          for category in variable.categories:
            space.add(
                vz.ParameterConfig.factory(
                    name=f'{variable.name}.ohe._{category}', bounds=(0., 1.)))
        else:
          raise ValueError(f'Unknown policy: {self._categorical_policy}')
      elif variable.optional:
        space.add(variable.as_parameter_config())
        # For optional parameters, handle the na dimension.
        if self._na_policy == NaPolicy.DROP:
          pass
        elif self._na_policy == NaPolicy.DISCRETE:
          # Optional variable.
          space.add(
              vz.ParameterConfig.factory(
                  f'{variable.name}.na', feasible_values=(0., 1.)))
        elif self._na_policy == NaPolicy.CONTINUOUS:
          # Optional variable.
          space.add(
              vz.ParameterConfig.factory(
                  f'{variable.name}.na', bounds=(0., 1.)))
        else:
          raise ValueError(f'Unknown policy: {self._na_policy}')
      elif data is not None:
        # TODO: Support integer conversions
        # TODO: Clean up with PEP 572 once python 3.8 arrives.
        uniq = np.unique(data.X[:, descriptor.column_index(variable.name)])
        if uniq.size < 10:
          space.add(
              variable.as_discrete_parameter_config(
                  list(variable.unscale(uniq))))
        else:
          space.add(variable.as_parameter_config())
      else:
        space.add(variable.as_parameter_config())

  def to_trials(self, dataset: _Dataset) -> List[vz.Trial]:
    """Convert HPOB data to Vizier trials."""
    xarray, yarray = dataset.X, dataset.Y
    trials = []
    for xrow, yrow in zip(xarray, yarray):
      trial = vz.Trial()
      params = trial.parameters
      for val, column_name in zip(xrow, self._descriptor.order):
        if '.ohe._' in column_name:
          # Categorical parameter
          if self._categorical_policy == CategoricalPolicy.AS_CATEGORICAL:
            if val:
              variable_name, category = column_name.split('.ohe._')
              if variable_name in params:
                raise ValueError(
                    f'The categorical parmateter {variable_name} has '
                    'more than one-hot encoding dimensions set to non-zero: '
                    f'{params[variable_name]} and {category}.')
              params[variable_name] = category
          elif self._categorical_policy == CategoricalPolicy.AS_CONTINUOUS:
            params[column_name] = float(val)
          else:
            raise ValueError(f'Unknown policy: {self._categorical_policy}')
        elif column_name.endswith('.na'):
          if self._na_policy == NaPolicy.DROP:
            # Delete the associated parameter.
            variable_name = column_name[:column_name.find('.na')]
            if variable_name not in params:
              raise ValueError(
                  f'This code assumes that VARIABLE column precedes VARIABLE.na,'
                  f'which is the case for HPOB-v3. '
                  f'However, {variable_name} did not precede {column_name})')
            if val:
              del params[variable_name]
          elif self._na_policy in (NaPolicy.DISCRETE, NaPolicy.CONTINUOUS):
            # Treat .na column as a regular parameter.
            params[column_name] = float(val)
          else:
            raise ValueError(f'Unknown policy: {self._na_policy}')
        else:
          params[column_name] = self._descriptor.variables[column_name].unscale(  # pytype: disable=unsupported-operands
              val)
      trial.complete(vz.Measurement({METRIC_NAME: yrow}))
      trials.append(trial)

    return trials

  def array_dim(self) -> int:
    """Returns the second dimension of the to_array()."""
    return len(self._descriptor.order)

  def to_array(self, trials: Sequence[vz.Trial]) -> np.ndarray:
    """Convert trial parameters to HPOB scaled array.

    Args:
      trials: Length N sequence.

    Returns:
      Array of shape [N, D], where D = self.array_dim()
    """
    all_values = []
    for trial in trials:
      params = trial.parameters
      values = []
      for column_name in self._descriptor.order:
        if '.ohe._' in column_name:
          if self._categorical_policy == CategoricalPolicy.AS_CATEGORICAL:
            # one hot embedding
            variable_name, category = column_name.split('.ohe._')
            value = float(params.get_value(variable_name) == category)
          elif self._categorical_policy == CategoricalPolicy.AS_CONTINUOUS:
            # continuous
            value = float(params.get_value(column_name))
          else:
            raise ValueError(f'Unknown policy: {self._categorical_policy}')
        elif column_name.endswith('.na'):
          if self._na_policy == NaPolicy.DROP:
            variable_name = column_name[:column_name.find('.na')]
            value = float(params.get_value(variable_name, None) is None)
          elif self._na_policy in (NaPolicy.CONTINUOUS, NaPolicy.DISCRETE):
            value = float(params.get_value(column_name))
          else:
            raise ValueError(f'Unknown policy: {self._na_policy}')
        else:
          if column_name in params:
            value = params.get_value(column_name)
            value = self._descriptor.variables[column_name].scale(value)  # pytype: disable=wrong-arg-types
          else:
            value = .0
        values.append(value)
      all_values.append(values)

    return np.array(all_values, dtype=np.float64)


def _surrogate_bounds(handler: handler_lib.HPOBHandler, search_space_id: str,
                      dataset_id: str) -> Tuple[float, float]:
  """Surrogate function bounds."""
  surrogate_name = 'surrogate-' + search_space_id + '-' + dataset_id
  y_min = handler.surrogates_stats[surrogate_name]['y_min']
  y_max = handler.surrogates_stats[surrogate_name]['y_max']
  return y_min, y_max


@attr.define
class HPOBExperimenter(experimenter.Experimenter):
  """HPOB Experimenter. Use HPOBContainer to create a valid instance."""
  _converter: _HPOBVizierConverter = attr.field()
  _handler: handler_lib.HPOBHandler = attr.field()

  _search_space_id: str = attr.field()
  _dataset_id: str = attr.field()
  _seed_trials: Sequence[vz.Trial] = attr.field()
  _normalize_y: bool = attr.field(kw_only=True)

  # Surrogate maps scaled and log-transformed features to raw objective values.
  _surrogate: xgb.Booster = attr.field(init=False)
  # Minimum value of the EvaluateContinuous output.
  _y_min: float = attr.field(init=False)
  # Maximum value of the EvaluateContinuous output.
  _y_max: float = attr.field(init=False)

  def __attrs_post_init__(self):
    self._surrogate = self._load_surrogate(self._search_space_id,
                                           self._dataset_id)
    self._y_min, self._y_max = _surrogate_bounds(self._handler,
                                                 self._search_space_id,
                                                 self._dataset_id)

  def _load_surrogate(self, search_space_id: str,
                      dataset_id: str) -> xgb.Booster:
    bst_surrogate = xgb.Booster()
    surrogate_name = 'surrogate-' + search_space_id + '-' + dataset_id
    bst_surrogate.load_model(self._handler.surrogates_dir + surrogate_name +
                             '.json')
    return bst_surrogate

  def array_dim(self) -> int:
    return self._converter.array_dim()

  def normalize_ys(self, y_array: np.ndarray) -> np.ndarray:
    new_y = (y_array - self._y_min) / (self._y_max - self._y_min)
    # Clip to eliminate potential numerical errors.
    new_y = np.clip(new_y, 0., 1.)
    return new_y

  def unnormalize_ys(self, y_array: np.ndarray) -> np.ndarray:
    new_y = self._y_min + y_array * (self._y_max - self._y_min)
    # Clip to eliminate potential numerical errors.
    new_y = np.clip(new_y, self._y_min, self._y_max)
    return new_y

  def _EvaluateArray(self, x_array: np.ndarray, normalize: bool) -> np.ndarray:
    x_q = xgb.DMatrix(x_array)
    new_y = self._surrogate.predict(x_q)
    if normalize:
      return self.normalize_ys(new_y)
    else:
      return np.clip(new_y, self._y_min, self._y_max)

  def EvaluateArray(self, x_array: np.ndarray) -> np.ndarray:
    """Evaluates the surrogate model on numpy array.

    Args:
      x_array: Size [N, D] array where D = self.array_dim(). The entries are in
        [0, 1] range.

    Returns:
      Size [N] array.
    """
    return self._EvaluateArray(x_array, self._normalize_y)

  def _EvaluateContinuous(self, trial: vz.Trial, normalize: bool) -> float:
    x_array = self._converter.to_array([trial])
    y_array = self._EvaluateArray(x_array[0].reshape(1, -1), normalize)
    return float(y_array.item())

  def EvaluateContinuous(self, trial: vz.Trial) -> float:
    """Returns a float."""
    # Convert Trial to X matrix.
    return self._EvaluateContinuous(trial, normalize=self._normalize_y)

  def evaluate(self, suggestions: Sequence[vz.Trial]):
    """Populates three metrics."""
    for suggestion in suggestions:
      unnormalized = self._EvaluateContinuous(suggestion, normalize=False)
      normalized = self._EvaluateContinuous(suggestion, normalize=True)
      y = normalized if self._normalize_y else unnormalized

      final_measurement = vz.Measurement(
          metrics={
              METRIC_NAME: y,
              f'{METRIC_NAME}_unnormalized': unnormalized,
              f'{METRIC_NAME}_normalized': normalized
          })
      suggestion.complete(final_measurement)

  def problem_statement(self) -> vz.ProblemStatement:
    problem = copy.deepcopy(self._converter.problem)
    problem.metadata.ns('hpob').update({
        'search_space_id': self._search_space_id,
        'dataset_id': self._dataset_id
    })
    return problem

  def __str__(self) -> str:
    return (f'HPOB experimenter on search_space_id:{self._search_space_id}, '
            f'dataset_id:{self._dataset_id}')

  def DebugString(self) -> str:
    return str(self)

  def get_initial_trials(
      self,
      count: int = 5,
      *,
      hpob_seed: Optional[str] = None,
      rng: Optional[np.random.RandomState] = None) -> List[vz.Trial]:
    """Get an initial batch of trials.

    Args:
      count:
      hpob_seed: If set, must be one of ( 'test0', 'test1', 'test2', 'test3',
        'test4'). Uses the fixed seed trials from HPO-B.
      rng: Must be set if hpob_seed is unset. Chooses initial trials randomly.

    Returns:
      List of seed trials.
    """
    if hpob_seed:
      init_ids = self._handler.bo_initializations[self._search_space_id][
          self._dataset_id][hpob_seed][:count]
    elif rng is None:
      raise ValueError('rng must be provided if hpob_seed is not set.')
    else:
      init_ids = rng.choice(
          np.arange(len(self._seed_trials)), count, replace=False)

    initial_trials = [self._seed_trials[init_id] for init_id in init_ids]
    # Assign trial id and final_measurement.objective_value.
    for i, trial in enumerate(initial_trials):
      trial.id = i + 1
      for metric in trial.final_measurement.metrics:
        if metric.name == METRIC_NAME:
          break
      else:
        raise ValueError(f'No metric named {METRIC_NAME} is found.')
    return initial_trials


# Download files from https://github.com/releaunifreiburg/HPO-B/.
ROOT_DIR = 'hpob-data/'
SURROGATES_DIR = 'saved-surrogates/'

DEFAULT_TEST_MODE = 'v3-test'
DEFAULT_MODE = 'v3-train-augmented'

TRAIN = 'train'
VALID = 'validation'
TEST = 'test'

SPLITS = [TRAIN, VALID, TEST]


class HPOBContainer:
  """HPOB container."""

  def __init__(
      self,
      handler: Optional[handler_lib.HPOBHandler] = None,
      *,
      auto_discretize: bool = False,
      na_policy: NaPolicy = NaPolicy.DROP,
      categorical_policy: CategoricalPolicy = CategoricalPolicy.AS_CATEGORICAL,
      normalize_y: bool = False,
      clip_y: bool = True,
      metadata_dir: str = ROOT_DIR,
      use_surrogate_values: bool = True):
    """Init.

    Args:
      handler: HPOBHandler. If not provided, a new one is automatically created
        from the default path.
      auto_discretize: If True, inspect the data columns and identify discrete
        parameters.
      na_policy: See NaPolicy.
      categorical_policy: See CategoricalPolicy.
      normalize_y: Normalize y values to [0, 1] from handler's surrogates_stats.
      clip_y: clip y value by [y_min, y_max] from handler's surrogates_stats.
        This is required to prepare initial seeds in the continuous evaluation
        setting.
      metadata_dir: Path to `meta_dataset_descriptors_v2.json` file.
      use_surrogate_values: If True, use surrogate values in get_study() and
        get_xy() methods.
    """
    if not handler:
      handler = handler_lib.HPOBHandler(
          root_dir=ROOT_DIR, mode=DEFAULT_MODE, surrogates_dir=SURROGATES_DIR)
    metadata_path = metadata_dir + 'meta_dataset_descriptors_v2.json'
    with Open(metadata_path, 'rt') as f:
      self._meta_dataset_descriptors_v2 = json.load(f)
    self._handler = handler

    self._auto_discretize = auto_discretize
    self._na_policy = na_policy
    self._categorical_policy = categorical_policy
    self._clip_y = clip_y
    self._normalize_y = normalize_y
    self._use_surrogate_values = use_surrogate_values

    self._all_data: Dict[Tuple[str, str], _Dataset] = dict()
    for split in (self._handler.meta_train_data,
                  self._handler.meta_validation_data,
                  self._handler.meta_test_data):
      if split is not None:
        for search_space_id, datasets in split.items():
          for dataset_id, datadict in datasets.items():
            y = datadict['y']
            if self._clip_y:
              y_min, y_max = _surrogate_bounds(self._handler, search_space_id,
                                               dataset_id)
              y = np.clip(y, y_min, y_max)
            self._all_data[search_space_id,
                           dataset_id] = _Dataset(datadict['X'], y)

  def dataset_keys(self, split: str) -> Iterable[Tuple[str, str]]:
    """Iterate through all dataset keys for the given split.

    Args:
      split: HPOBContainer.TRAIN, VALID, or TEST.

    Yields:
      Tuple of search_space_id and dataset_id.
    """
    if split == TRAIN:
      root = self._handler.meta_train_data
    elif split == VALID:
      root = self._handler.meta_validation_data
    elif split == TEST:
      root = self._handler.meta_test_data

    if root is not None:
      for search_space_id, dataset_ids in root.items():
        for dataset_id in dataset_ids:
          yield search_space_id, dataset_id

  def _get_descriptor(self, search_space_id: str) -> _SearchspaceDescriptor:
    """Get the search space descriptor."""
    d = self._meta_dataset_descriptors_v2[search_space_id]
    variables = d['variables']
    descriptors = dict()
    for name in variables:
      if name.endswith('.na'):
        # These variables indicate whether an "optional" parameter is specified
        # or not. We ignore them for now.
        continue
      descriptors[name] = _VariableDescriptor.from_info(
          name,
          variables[name],
          apply_log=name in d['variables_to_apply_log'],
          optional=f'{name}.na' in variables)
    return _SearchspaceDescriptor(descriptors, d['variables_order'])

  def get_xy(self, search_space_id: str,
             dataset_id: str) -> Tuple[np.ndarray, np.ndarray]:
    data, _ = self._get_data_and_converter(search_space_id, dataset_id)
    return data.x, data.y  # pytype: disable=attribute-error

  def _get_data_and_converter(
      self, search_space_id: str,
      dataset_id: str) -> Tuple[_Dataset, _HPOBVizierConverter]:
    """Get data and converter."""
    data: _Dataset = copy.deepcopy(self._all_data[search_space_id, dataset_id])

    if self._normalize_y:
      y_min, y_max = _surrogate_bounds(self._handler, search_space_id,
                                       dataset_id)
      data.Y = (data.Y - y_min) / (y_max - y_min)

    converter = _HPOBVizierConverter(
        self._get_descriptor(search_space_id),
        (data if self._auto_discretize else None),
        na_policy=self._na_policy,
        categorical_policy=self._categorical_policy)
    if self._use_surrogate_values:
      surrogate = self._get_experimenter_from_data_and_converter(
          data, converter, search_space_id, dataset_id)
      data.Y = surrogate.EvaluateArray(data.X)
    return data, converter

  def _get_experimenter_from_data_and_converter(
      self, data: _Dataset, converter: _HPOBVizierConverter,
      search_space_id: str, dataset_id: str) -> HPOBExperimenter:
    return HPOBExperimenter(
        converter,
        self._handler,
        search_space_id,
        dataset_id,
        converter.to_trials(data),
        normalize_y=self._normalize_y)

  def get_experimenter(self, search_space_id: str,
                       dataset_id: str) -> HPOBExperimenter:
    data, converter = self._get_data_and_converter(search_space_id, dataset_id)
    return self._get_experimenter_from_data_and_converter(
        data, converter, search_space_id, dataset_id)

  def get_problem_and_trials(self, search_space_id: str,
                             dataset_id: str) -> vz.ProblemAndTrials:
    data, converter = self._get_data_and_converter(search_space_id, dataset_id)
    return vz.ProblemAndTrials(
        problem=converter.problem, trials=converter.to_trials(data))



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for experimenter_factory."""

import numpy as np
from vizier import pyvizier
from vizier._src.benchmarks.experimenters import experimenter_factory

from absl.testing import absltest
from absl.testing import parameterized


class ExperimenterFactoryTest(parameterized.TestCase):

  @parameterized.parameters(
      {'bbob_name': 'Sphere'},
      {'bbob_name': 'LinearSlope'},
      {'bbob_name': 'RosenbrockRotated'},
      {'bbob_name': 'SchaffersF7IllConditioned'},
  )
  def testBBOBFactory(self, bbob_name):
    dim = 4
    bbob_factory = experimenter_factory.BBOBExperimenterFactory(
        name=bbob_name, dim=dim)
    exptr = bbob_factory()

    parameters = exptr.problem_statement().search_space.parameters
    self.assertLen(parameters, dim)

    t = pyvizier.Trial(parameters={
        param.name: float(index) for index, param in enumerate(parameters)
    })
    exptr.evaluate([t])
    self.assertEqual(t.status, pyvizier.TrialStatus.COMPLETED)

  def testBBOBFactoryError(self):

    with self.assertRaisesRegex(ValueError, 'not a valid BBOB'):
      experimenter_factory.BBOBExperimenterFactory(name='Error', dim=3)()

  def testSingleObjectiveFactory(self):
    dim = 5
    bbob_factory = experimenter_factory.BBOBExperimenterFactory(
        name='Sphere', dim=dim)
    exptr_factory = experimenter_factory.SingleObjectiveExperimenterFactory(
        base_factory=bbob_factory,
        shift=np.asarray(1.9),
        noise_type='moderate_gaussian')
    exptr = exptr_factory()

    self.assertIn('Shifting', str(exptr))
    self.assertIn('Noisy', str(exptr))

    parameters = exptr.problem_statement().search_space.parameters
    self.assertLen(parameters, dim)

    t = pyvizier.Trial(parameters={
        param.name: float(index) for index, param in enumerate(parameters)
    })
    exptr.evaluate([t])
    self.assertEqual(t.status, pyvizier.TrialStatus.COMPLETED)

  def testSingleObjectiveFactoryDiscrete(self):
    dim = 5
    bbob_factory = experimenter_factory.BBOBExperimenterFactory(
        name='Sphere', dim=dim
    )
    exptr_factory = experimenter_factory.SingleObjectiveExperimenterFactory(
        base_factory=bbob_factory,
        shift=np.asarray(1.9),
        noise_type='moderate_gaussian',
        num_normalization_samples=10,
        discrete_dict={0: 3, 1: 5},
        categorical_dict={4: 4},
    )
    exptr = exptr_factory()

    self.assertIn('Shifting', str(exptr))
    self.assertIn('Noisy', str(exptr))
    self.assertIn('Discretizing', str(exptr))
    self.assertIn('Normalizing', str(exptr))

    space = exptr.problem_statement().search_space
    self.assertEqual(space.num_parameters(pyvizier.ParameterType.DISCRETE), 2)
    self.assertEqual(
        space.num_parameters(pyvizier.ParameterType.CATEGORICAL), 1
    )

    parameters = {}
    for param in space.parameters:
      if param.type == pyvizier.ParameterType.DOUBLE:
        parameters[param.name] = param.bounds[0]
      else:
        parameters[param.name] = param.feasible_values[0]
    t = pyvizier.Trial(parameters=parameters)
    exptr.evaluate([t])
    self.assertEqual(t.status, pyvizier.TrialStatus.COMPLETED)

  def testSingleObjectiveFactoryError(self):
    dim = 4
    bbob_factory = experimenter_factory.BBOBExperimenterFactory(
        name='Sphere', dim=dim)

    with self.assertRaisesRegex(ValueError, 'not supported'):
      experimenter_factory.SingleObjectiveExperimenterFactory(
          base_factory=bbob_factory, shift=np.asarray(1.9),
          noise_type='ERROR')()

  def testDiscreteSingleObjectiveFactoryError(self):
    dim = 4
    bbob_factory = experimenter_factory.BBOBExperimenterFactory(
        name='Sphere', dim=dim
    )

    with self.assertRaisesRegex(ValueError, 'overlap'):
      experimenter_factory.SingleObjectiveExperimenterFactory(
          base_factory=bbob_factory,
          shift=np.asarray(1.9),
          noise_type='moderate_gaussian',
          num_normalization_samples=10,
          discrete_dict={0: 3, 1: 5},
          categorical_dict={1: 4},
      )()

  def testSingleObjectiveSerialization(self):
    dim = 5
    bbob_factory = experimenter_factory.BBOBExperimenterFactory(
        name='Sphere', dim=dim
    )
    bbob_metadata = bbob_factory.dump()
    recovered_bbob = experimenter_factory.BBOBExperimenterFactory.recover(
        bbob_metadata
    )
    self.assertEqual(recovered_bbob.dim, dim)
    self.assertEqual(recovered_bbob.name, 'Sphere')

    exptr_factory = experimenter_factory.SingleObjectiveExperimenterFactory(
        base_factory=bbob_factory,
        shift=np.asarray(1.9),
        noise_type='moderate_gaussian',
        num_normalization_samples=10,
        discrete_dict={0: 3, 1: 5},
    )
    factory_metadata = exptr_factory.dump()
    recovered_factory = (
        experimenter_factory.SingleObjectiveExperimenterFactory.recover(
            factory_metadata
        )
    )
    self.assertEqual(recovered_factory.noise_type, 'moderate_gaussian')
    self.assertEqual(recovered_factory.num_normalization_samples, 10)
    np.testing.assert_array_equal(recovered_factory.shift, np.asarray(1.9))
    self.assertDictEqual(recovered_factory.discrete_dict, {0: 3, 1: 5})


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for vizier.pyvizier.shared.trial."""
import copy
import datetime

from typing import Sequence

import numpy as np

from vizier._src.pyvizier.shared import trial
from absl.testing import absltest
from absl.testing import parameterized

Metric = trial.Metric
Measurement = trial.Measurement


class MetricTest(absltest.TestCase):

  def testMetricCreation(self):
    _ = Metric(value=0, std=0.5)

  def testMetricCannotHaveNaN(self):
    with self.assertRaises(ValueError):
      _ = Metric(value=np.nan, std=-np.nan)

  def testMetricCannotHaveNegativeStd(self):
    with self.assertRaises(ValueError):
      _ = Metric(value=0, std=-0.5)


class MeasurementTest(absltest.TestCase):

  def testMetricsInitializedFromFloats(self):
    m = Measurement()
    m.metrics = dict(a=0.3)
    self.assertEqual(m.metrics['a'], Metric(0.3))
    m.metrics['b'] = 0.5
    self.assertEqual(m.metrics, {'a': Metric(0.3), 'b': Metric(0.5)})

  def testMetrics(self):
    m = Measurement()
    m.metrics = dict(a=Metric(0.3))
    self.assertEqual(m.metrics['a'], Metric(0.3))

  def testTimeStampsAreNotFrozen(self):
    m = Measurement()
    m.elapsed_secs = 1.0
    m.steps = 5


ParameterValue = trial.ParameterValue


class ParameterValueTest(parameterized.TestCase):

  @parameterized.parameters((True,), (False,))
  def testBool(self, bool_value):
    value = ParameterValue(bool_value)
    self.assertEqual(value.as_float, float(bool_value))
    self.assertEqual(value.as_int, int(bool_value))
    self.assertEqual(value.as_str, str(bool_value))

  def testIntegralFloat0(self):
    value = ParameterValue(0.0)
    self.assertEqual(value.as_float, 0.0)
    self.assertEqual(value.as_int, 0)
    self.assertEqual(value.as_bool, False)
    self.assertEqual(value.as_str, '0.0')

  def testIntegralFloat1(self):
    value = ParameterValue(1.0)
    self.assertEqual(value.as_float, 1.0)
    self.assertEqual(value.as_int, 1)
    self.assertEqual(value.as_bool, True)
    self.assertEqual(value.as_str, '1.0')

  def testIntegralFloat2(self):
    value = ParameterValue(2.0)
    self.assertEqual(value.as_float, 2.0)
    self.assertEqual(value.as_int, 2)
    self.assertIsNone(value.as_bool)
    self.assertEqual(value.as_str, '2.0')

  def testInteger0(self):
    value = ParameterValue(0)
    self.assertEqual(value.as_float, 0)
    self.assertEqual(value.as_int, 0)
    self.assertEqual(value.as_bool, False)
    self.assertEqual(value.as_str, '0')

  def testInteger1(self):
    value = ParameterValue(1)
    self.assertEqual(value.as_float, 1)
    self.assertEqual(value.as_int, 1)
    self.assertEqual(value.as_bool, True)
    self.assertEqual(value.as_str, '1')

  def testInteger2(self):
    value = ParameterValue(2)
    self.assertEqual(value.as_float, 2)
    self.assertEqual(value.as_int, 2)
    self.assertIsNone(value.as_bool)
    self.assertEqual(value.as_str, '2')

  def testStringTrue(self):
    value = ParameterValue(trial.TRUE_VALUE)
    self.assertEqual(value.as_bool, True)
    self.assertEqual(value.as_str, trial.TRUE_VALUE)

  def testStringFalse(self):
    value = ParameterValue(trial.FALSE_VALUE)
    self.assertEqual(value.as_bool, False)
    self.assertEqual(value.as_str, trial.FALSE_VALUE)

  def testStringFloat1(self):
    value = ParameterValue('1.0')
    self.assertEqual(value.as_float, 1.0)
    self.assertIsNone(value.as_int)
    self.assertIsNone(value.as_bool)
    self.assertEqual(value.as_str, '1.0')

  def testStringInt1(self):
    value = ParameterValue('1')
    self.assertEqual(value.as_float, 1.0)
    self.assertEqual(value.as_int, 1)
    self.assertIsNone(value.as_bool)
    self.assertEqual(value.as_str, '1')

  def testParameterCanHaveNonFiniteValues(self):
    ParameterValue(float('nan'))
    ParameterValue(value=float('inf'))
    ParameterValue(value=float('inf'))


class TrialTest(absltest.TestCase):

  def testCompleteInplace(self):
    test = trial.Trial()
    measurement = Measurement(metrics={
        'pr-auc': Metric(value=0.8),
        'latency': Metric(value=32)
    })
    completed = test.complete(measurement, inplace=True)

    # The trial was completed in place.
    self.assertEqual(test.final_measurement, measurement)
    self.assertLessEqual(test.completion_time,
                         datetime.datetime.now().astimezone())
    self.assertGreaterEqual(test.completion_time, test.creation_time)
    self.assertGreaterEqual(test.duration.total_seconds(), 0)

    self.assertEqual(completed.final_measurement, measurement)
    self.assertLessEqual(completed.completion_time,
                         datetime.datetime.now().astimezone())
    self.assertGreaterEqual(completed.completion_time, completed.creation_time)
    self.assertGreaterEqual(completed.duration.total_seconds(), 0)

    # completed is the same reference as test.
    self.assertEqual(test, completed)

  def testCompleteNotInplace(self):
    """Complete with inplace=False."""
    test = trial.Trial()
    measurement = Measurement(metrics={
        'pr-auc': Metric(value=0.8),
        'latency': Metric(value=32)
    })

    test_copy = copy.deepcopy(test)

    completed = test.complete(measurement, inplace=False)

    # The returned Trial is completed.
    self.assertEqual(completed.final_measurement, measurement)
    self.assertGreaterEqual(completed.completion_time, completed.creation_time)
    self.assertLessEqual(completed.completion_time,
                         datetime.datetime.now().astimezone())
    self.assertGreaterEqual(completed.duration.total_seconds(), 0)
    self.assertEqual(completed.status, trial.TrialStatus.COMPLETED)
    self.assertTrue(completed.is_completed)

    # The original Trial is unchanged.
    self.assertEqual(test_copy, test)
    self.assertIsNone(test.final_measurement)
    self.assertIsNone(test.completion_time)
    self.assertIsNone(test.duration)
    self.assertEqual(test.status, trial.TrialStatus.ACTIVE)
    self.assertFalse(test.is_completed)

  def testCompleteInfeasible(self):
    test = trial.Trial()
    measurement = Measurement(metrics={
        'pr-auc': Metric(value=0.8),
        'latency': Metric(value=32)
    })
    completed = test.complete(
        measurement, inplace=False, infeasibility_reason='reason')
    # Test infeasibility.
    self.assertTrue(completed.infeasible)
    self.assertEqual(completed.infeasibility_reason, 'reason')

  def testCompleteInfeasible2(self):
    test = trial.Trial(infeasibility_reason='reason')
    measurement = Measurement(metrics={
        'pr-auc': Metric(value=0.8),
        'latency': Metric(value=32)
    })
    completed = test.complete(measurement, inplace=False)
    # When infeasibility not provided, the trial shoud remain infeasible.
    self.assertTrue(completed.infeasible)
    self.assertEqual(completed.infeasibility_reason, 'reason')

  def testCompleteInfeasible3(self):
    test = trial.Trial(infeasibility_reason='reason')
    measurement = Measurement(metrics={
        'pr-auc': Metric(value=0.8),
        'latency': Metric(value=32)
    })
    completed = test.complete(
        measurement, inplace=False, infeasibility_reason='other')
    # Infeasibility reason should be updated.
    self.assertTrue(completed.infeasible)
    self.assertEqual(completed.infeasibility_reason, 'other')
    # The original trial should retain the original reason.
    self.assertTrue(test.infeasible)
    self.assertEqual(test.infeasibility_reason, 'reason')

  def testCompleteEmptyInfeasible(self):
    test = trial.Trial()
    measurement = Measurement(metrics={
        'pr-auc': Metric(value=0.8),
        'latency': Metric(value=32)
    })
    completed = test.complete(
        measurement, inplace=False, infeasibility_reason='')
    # Test infeasibility.
    self.assertTrue(completed.infeasible)
    self.assertEqual(completed.infeasibility_reason, '')

  def testCompleteInfeasibleInplace(self):
    test = trial.Trial()
    measurement = Measurement(metrics={
        'pr-auc': Metric(value=0.8),
        'latency': Metric(value=32)
    })
    test.complete(measurement, inplace=True, infeasibility_reason='reason')
    # Test infeasibility.
    self.assertTrue(test.infeasible)
    self.assertEqual(test.infeasibility_reason, 'reason')

  def testCompleteInfeasibleInplace2(self):
    test = trial.Trial(infeasibility_reason='reason')
    measurement = Measurement(metrics={
        'pr-auc': Metric(value=0.8),
        'latency': Metric(value=32)
    })
    test.complete(measurement, inplace=True)
    # When infeasibility not provided, the trial shoud remain infeasible.
    self.assertTrue(test.infeasible)
    self.assertEqual(test.infeasibility_reason, 'reason')

  def testCompleteInfeasibleInplace3(self):
    test = trial.Trial(infeasibility_reason='reason')
    measurement = Measurement(metrics={
        'pr-auc': Metric(value=0.8),
        'latency': Metric(value=32)
    })
    test.complete(measurement, inplace=True, infeasibility_reason='other')
    # Infeasibility reason should be updated.
    self.assertTrue(test.infeasible)
    self.assertEqual(test.infeasibility_reason, 'other')

  def testCompleteEmptyInfeasibleInplace(self):
    test = trial.Trial()
    measurement = Measurement(metrics={
        'pr-auc': Metric(value=0.8),
        'latency': Metric(value=32)
    })
    completed = test.complete(
        measurement, inplace=True, infeasibility_reason='')
    # Test infeasibility.
    self.assertTrue(completed.infeasible)
    self.assertEqual(completed.infeasibility_reason, '')

  def testDefaultsNotShared(self):
    """Make sure default parameters are not shared between instances."""
    trial1 = trial.Trial()
    trial2 = trial.Trial()
    trial1.parameters['x1'] = trial.ParameterValue(5)
    self.assertEmpty(trial2.parameters)

  def testCreationTime(self):
    trial1 = trial.Trial()
    trial2 = trial.Trial()
    self.assertGreater(trial2.creation_time, trial1.creation_time)

  def testCompletionTime(self):
    trial1 = trial.Trial(
        final_measurement=trial.Measurement(
            metrics={'pr-auc': Metric(value=0.8)}
        )
    )
    self.assertEqual(trial1.creation_time, trial1.completion_time)
    trial2 = trial.Trial(infeasibility_reason='reasons')
    self.assertEqual(trial2.creation_time, trial2.completion_time)


class ParameterDictTest(parameterized.TestCase):

  @parameterized.parameters((True,), (3,), (1.,), ('aa',))
  def testAssignRawValue(self, v):
    d = trial.ParameterDict()
    d['p1'] = v
    self.assertEqual(d.get('p1'), trial.ParameterValue(v))
    self.assertEqual(d.get_value('p1'), v)
    self.assertEqual(d.get_value('p2', 'default'), 'default')
    self.assertLen(d, 1)
    self.assertLen(d.items(), 1)

  @parameterized.parameters((True,), (3,), (1.,), ('aa',))
  def testAssignWrappedValue(self, v):
    d = trial.ParameterDict()
    v = trial.ParameterValue(v)
    d['p1'] = v
    self.assertEqual(d.get('p1'), v)
    self.assertEqual(d.get_value('p1'), v.value)
    self.assertEqual(d.get_value('p2', 'default'), 'default')
    self.assertLen(d, 1)
    self.assertLen(d.items(), 1)


class SuggestionTestI(absltest.TestCase):

  def testToTrial(self):
    suggestion = trial.TrialSuggestion({'a': 3, 'b': True})
    suggestion.metadata['key'] = 'value'

    t = suggestion.to_trial(1)
    self.assertEqual(t.id, 1)
    self.assertEqual(t.parameters, suggestion.parameters)
    self.assertEqual(t.metadata, suggestion.metadata)


class TrialFilterTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(filtr=trial.TrialFilter(), answers=[True, True, True, True]),
      dict(
          filtr=trial.TrialFilter(ids=(2, 3)),
          answers=[False, True, True, False]),
      dict(
          filtr=trial.TrialFilter(min_id=3, ids=(2, 3)),
          answers=[False, False, True, False]),
      dict(
          filtr=trial.TrialFilter(
              min_id=2,
              max_id=3,
              ids=(1, 2, 3, 4),
              status=[trial.TrialStatus.REQUESTED]),
          answers=[False, True, False, False]))
  def test_filter(self, filtr: trial.TrialFilter, answers: Sequence[bool]):
    trials = (
        trial.Trial(id=1),  # ACTIVE
        trial.Trial(id=2, is_requested=True),  #  REQUESTED
        trial.Trial(id=3, stopping_reason='stopping'),  # STOPPING
        trial.Trial(id=4).complete(trial.Measurement()),  # COMPLETED
    )
    self.assertSequenceEqual([filtr(t) for t in trials], answers)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Convergence test for Eagle Auto Tuner."""

from typing import Optional
import attrs
from vizier import benchmarks
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import random
from vizier._src.algorithms.designers.eagle_strategy import eagle_strategy
from vizier._src.algorithms.designers.eagle_strategy import testing
from vizier._src.algorithms.designers.meta_learning import eagle_meta_learning
from vizier._src.algorithms.designers.meta_learning import meta_learning
from vizier._src.algorithms.testing import comparator_runner
from vizier._src.benchmarks.experimenters.synthetic import bbob
from absl.testing import absltest
from absl.testing import parameterized


def _eagle_designer_factory(
    problem: vz.ProblemStatement, seed: Optional[int], **kwargs
):
  """Creates an EagleStrategyDesigner with hyper-parameters and seed."""
  config = eagle_strategy.FireflyAlgorithmConfig()
  # Unpack the hyperparameters into the Eagle config class.
  for param_name, param_value in kwargs.items():
    if param_name not in attrs.asdict(config):
      raise ValueError(f"'{param_name}' is not in FireflyAlgorithmConfig!")
    setattr(config, param_name, param_value)
  return eagle_strategy.EagleStrategyDesigner(
      problem_statement=problem,
      seed=seed,
      config=config,
  )


class EagleEagleMetaLearningConvergenceTest(parameterized.TestCase):
  """Convergence test for meta Eagle-Eagle designer.

  Note that all optimization problems are MINIMIZATION.
  """

  @parameterized.parameters(
      testing.create_continuous_exptr(bbob.Rastrigin),
      testing.create_categorical_exptr(),
  )
  def test_convergence(self, exptr):
    def _random_designer_factory(problem, seed):
      return random.RandomDesigner(problem.search_space, seed=seed)

    def _meta_eagle_eagle_designer_factory(problem, seed):
      meta_config = meta_learning.MetaLearningConfig(
          num_trials_per_tuning=100,
          tuning_max_num_trials=1000,
          tuning_min_num_trials=200,
      )
      return meta_learning.MetaLearningDesigner(
          problem=problem,
          tuned_designer_factory=_eagle_designer_factory,
          meta_designer_factory=_eagle_designer_factory,
          tuning_hyperparams=eagle_meta_learning.meta_eagle_search_space(),
          config=meta_config,
          seed=seed,
      )

    random_benchmark_state_factory = benchmarks.DesignerBenchmarkStateFactory(
        designer_factory=_random_designer_factory, experimenter=exptr
    )

    meta_benchmark_state_factory = benchmarks.DesignerBenchmarkStateFactory(
        designer_factory=_meta_eagle_eagle_designer_factory,
        experimenter=exptr,
    )
    evaluations = 1_500
    # Random designer batch size is large to expedite run time.
    comparator_runner.SimpleRegretComparisonTester(
        baseline_num_trials=2 * evaluations,
        candidate_num_trials=evaluations,
        baseline_suggestion_batch_size=2 * evaluations,
        candidate_suggestion_batch_size=5,
        baseline_num_repeats=5,
        candidate_num_repeats=1,
        alpha=0.05,
        goal=vz.ObjectiveMetricGoal.MINIMIZE,
    ).assert_benchmark_state_better_simple_regret(
        baseline_benchmark_state_factory=random_benchmark_state_factory,
        candidate_benchmark_state_factory=meta_benchmark_state_factory,
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Categorical benchmarks from https://github.com/QUVA-Lab/COMBO.

Lines with 'ATTENTION' mean that code was modified from original Github, in
order to fix bugs with the original code.
"""
from typing import List, Optional, Sequence, Tuple
import numpy as np

from vizier import pyvizier
from vizier._src.benchmarks.experimenters import experimenter
from vizier._src.benchmarks.experimenters.combo import common

fileOpen = open


class IsingExperimenter(experimenter.Experimenter):
  """Ising Sparisification Problem."""

  def __init__(self,
               lamda: float,
               ising_grid_h: int = 4,
               ising_grid_w: int = 4,
               ising_n_edges: int = 24,
               random_seed: Optional[int] = None):
    self._lamda = lamda
    self._ising_grid_h = ising_grid_h
    self._ising_grid_w = ising_grid_w
    self._ising_n_edges = ising_n_edges
    self._interaction = common.generate_ising_interaction(
        self._ising_grid_h, self._ising_grid_w, random_seed)
    self._covariance, self._partition_original = common.spin_covariance(
        self._interaction, (self._ising_grid_h, self._ising_grid_w))

    self._problem_statement = self.problem_statement()

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    for suggestion in suggestions:
      # TODO: Switch to using StudyConfig.
      x = np.array([
          int(suggestion.parameters[f'x_{i}'].value == 'True')
          for i in range(self._ising_n_edges)
      ])
      x_h, x_v = self._bocs_consistency_mapping(x)
      interaction_sparsified = x_h * self._interaction[
          0], x_v * self._interaction[1]
      log_partition_sparsified = common.log_partition(
          interaction_sparsified, (self._ising_grid_h, self._ising_grid_w))
      evaluation = common.ising_dense(
          ising_grid_h=self._ising_grid_h,
          interaction_original=self._interaction,
          interaction_sparsified=interaction_sparsified,
          covariance=self._covariance,
          log_partition_original=np.log(self._partition_original),
          log_partition_new=log_partition_sparsified)
      evaluation += self._lamda * float(np.sum(x))

      suggestion.complete(
          pyvizier.Measurement(metrics={
              self._problem_statement.single_objective_metric_name: evaluation
          }))

  def problem_statement(self) -> pyvizier.ProblemStatement:
    problem_statement = pyvizier.ProblemStatement()
    root = problem_statement.search_space.root
    for i in range(self._ising_n_edges):
      root.add_bool_param(name=f'x_{i}')
    problem_statement.metric_information.append(
        pyvizier.MetricInformation(
            name='main_objective', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE))
    return problem_statement

  def _bocs_consistency_mapping(self,
                                x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    horizontal_ind = [0, 2, 4, 7, 9, 11, 14, 16, 18, 21, 22, 23]
    vertical_ind = sorted(
        [elm for elm in range(24) if elm not in horizontal_ind])
    return x[horizontal_ind].reshape(
        (self._ising_grid_h, self._ising_grid_w - 1)), x[vertical_ind].reshape(
            (self._ising_grid_h - 1, self._ising_grid_w))


class ContaminationExperimenter(experimenter.Experimenter):
  """Contamination Control Problem."""

  def __init__(self,
               lamda: float,
               contamination_n_stages: int = 25,
               random_seed: Optional[int] = None):
    self._lamda = lamda
    self._contamination_n_stages = contamination_n_stages
    self._init_z, self._lambdas, self._gammas = self._generate_contamination_dynamics(
        random_seed)
    self._problem_statement = self.problem_statement()

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    for suggestion in suggestions:
      # TODO: Switch to using StudyConfig.
      x = np.array([
          int(suggestion.parameters[f'x_{i}'].value == 'True')
          for i in range(self._contamination_n_stages)
      ])
      evaluation = self._contamination(
          x=x,
          cost=np.ones(x.size),
          init_z=self._init_z,
          lambdas=self._lambdas,
          gammas=self._gammas,
          u=0.1,
          epsilon=0.05)
      evaluation += self._lamda * float(np.sum(x))
      suggestion.complete(
          pyvizier.Measurement(metrics={
              self._problem_statement.single_objective_metric_name: evaluation
          }))

  def problem_statement(self) -> pyvizier.ProblemStatement:
    problem_statement = pyvizier.ProblemStatement()
    root = problem_statement.search_space.root
    for i in range(self._contamination_n_stages):
      root.add_bool_param(name=f'x_{i}')
    problem_statement.metric_information.append(
        pyvizier.MetricInformation(
            name='main_objective', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE))
    return problem_statement

  def _contamination(self, x: np.ndarray, cost: np.ndarray, init_z: np.ndarray,
                     lambdas: np.ndarray, gammas: np.ndarray, u: float,
                     epsilon: float) -> float:
    assert x.size == self._contamination_n_stages

    rho = 1.0
    n_simulations = 100

    z = np.zeros((x.size, n_simulations))
    z[0] = lambdas[0] * (1.0 - x[0]) * (1.0 - init_z) + (
        1.0 - gammas[0] * x[0]) * init_z
    for i in range(1, self._contamination_n_stages):
      z[i] = lambdas[i] * (1.0 - x[i]) * (1.0 - z[i - 1]) + (
          1.0 - gammas[i] * x[i]) * z[i - 1]

    below_threshold = z < u
    constraints = np.mean(below_threshold, axis=1) - (1.0 - epsilon)

    return np.sum(x * cost - rho * constraints)

  def _generate_contamination_dynamics(
      self, random_seed=None) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    n_stages = self._contamination_n_stages
    n_simulations = 100

    init_alpha = 1.0
    init_beta = 30.0
    contam_alpha = 1.0
    contam_beta = 17.0 / 3.0
    restore_alpha = 1.0
    restore_beta = 3.0 / 7.0
    init_z = np.random.RandomState(random_seed).beta(
        init_alpha, init_beta, size=(n_simulations,))
    lambdas = np.random.RandomState(random_seed).beta(
        contam_alpha, contam_beta, size=(n_stages, n_simulations))
    gammas = np.random.RandomState(random_seed).beta(
        restore_alpha, restore_beta, size=(n_stages, n_simulations))

    return init_z, lambdas, gammas


class CentroidExperimenter(experimenter.Experimenter):
  """General parameter variant of the Ising Sparisification Problem."""

  def __init__(self,
               centroid_n_choice=3,
               centroid_grid=(4, 4),
               ising_grid_h: int = 4,
               random_seed: Optional[int] = None):
    self._centroid_n_choice = centroid_n_choice
    self._centroid_grid = centroid_grid
    self._centroid_n_edges = centroid_grid[0] * (centroid_grid[1] - 1) + (
        centroid_grid[0] - 1) * centroid_grid[1]
    self._ising_grid_h = ising_grid_h

    self._interaction_list = []
    self._covariance_list = []
    self._partition_original_list = []
    self._n_ising_models = 3
    ising_seeds = np.random.RandomState(random_seed).randint(
        0, 10000, (self._n_ising_models,))
    for i in range(self._n_ising_models):
      interaction = common.generate_ising_interaction(self._centroid_grid[0],
                                                      self._centroid_grid[1],
                                                      ising_seeds[i])
      covariance, partition_original = common.spin_covariance(
          interaction, self._centroid_grid)
      self._interaction_list.append(interaction)
      self._covariance_list.append(covariance)
      self._partition_original_list.append(partition_original)

    self._problem_statement = self.problem_statement()

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    for suggestion in suggestions:
      # TODO: Switch to using StudyConfig.
      x = np.array([
          int(suggestion.parameters[f'x_{i}'].value)
          for i in range(self._centroid_n_edges)
      ])
      interaction_mixed = self._edge_choice(x, self._interaction_list)
      log_partition_mixed = common.log_partition(
          interaction_mixed, self._centroid_grid)  # ATTENTION
      kld_sum = 0

      for i in range(self._n_ising_models):
        kld = common.ising_dense(
            ising_grid_h=self._centroid_grid[0],  # ATTENTION
            interaction_original=self._interaction_list[i],
            interaction_sparsified=interaction_mixed,
            covariance=self._covariance_list[i],
            log_partition_original=np.log(
                self._partition_original_list[i]),  # ATTENTION
            log_partition_new=log_partition_mixed)  # ATTENTION
        kld_sum += kld
      evaluation = float(kld_sum / float(self._n_ising_models))

      suggestion.complete(
          pyvizier.Measurement(metrics={
              self._problem_statement.single_objective_metric_name: evaluation
          }))

  def problem_statement(self) -> pyvizier.ProblemStatement:
    problem_statement = pyvizier.ProblemStatement()
    root = problem_statement.search_space.root
    for i in range(self._centroid_n_edges):
      root.add_categorical_param(
          name='x_{}'.format(i),
          feasible_values=[str(j) for j in range(self._centroid_n_choice)])
    problem_statement.metric_information.append(
        pyvizier.MetricInformation(
            name='main_objective', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE))
    return problem_statement

  def _edge_choice(
      self, x: np.ndarray, interaction_list: List[Tuple[np.ndarray, np.ndarray]]
  ) -> Tuple[np.ndarray, np.ndarray]:
    edge_weight = np.zeros(x.shape)
    for i in range(len(interaction_list)):
      edge_weight[x == i] = np.hstack([
          interaction_list[i][0].reshape(-1), interaction_list[i][1].reshape(-1)
      ])[x == i]
    grid_h, grid_w = self._centroid_grid
    split_ind = grid_h * (grid_w - 1)
    return edge_weight[:split_ind].reshape(
        (grid_h, grid_w - 1)), edge_weight[split_ind:].reshape(
            (grid_h - 1, grid_w))


class PestControlExperimenter(experimenter.Experimenter):
  """Pest Control Problem."""

  def __init__(self,
               pest_control_n_choice: int = 5,
               pest_control_n_stages: int = 25,
               random_seed: Optional[int] = None):
    self._pest_control_n_choice = pest_control_n_choice
    self._pest_control_n_stages = pest_control_n_stages
    self._random_seed = random_seed
    self._problem_statement = self.problem_statement()

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    for suggestion in suggestions:
      # TODO: Switch to using StudyConfig.
      x = np.array([
          int(suggestion.parameters[f'x_{i}'].value)
          for i in range(self._pest_control_n_stages)
      ])
      evaluation = self._pest_control_score(x)
      suggestion.complete(
          pyvizier.Measurement(metrics={
              self._problem_statement.single_objective_metric_name: evaluation
          }))

  def problem_statement(self) -> pyvizier.ProblemStatement:
    problem_statement = pyvizier.ProblemStatement()
    root = problem_statement.search_space.root
    for i in range(self._pest_control_n_stages):
      root.add_categorical_param(
          name='x_{}'.format(i),
          feasible_values=[str(j) for j in range(self._pest_control_n_choice)])
    problem_statement.metric_information.append(
        pyvizier.MetricInformation(
            name='main_objective', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE))
    return problem_statement

  def _pest_spread(self, curr_pest_frac: float, spread_rate: float,
                   control_rate: float, apply_control: bool):
    if apply_control:
      next_pest_frac = (1.0 - control_rate) * curr_pest_frac
    else:
      next_pest_frac = spread_rate * (1 - curr_pest_frac) + curr_pest_frac
    return next_pest_frac

  def _pest_control_score(self, x: np.ndarray) -> float:
    u = 0.1
    n_stages = x.size
    n_simulations = 100

    init_pest_frac_alpha = 1.0
    init_pest_frac_beta = 30.0
    spread_alpha = 1.0
    spread_beta = 17.0 / 3.0

    control_alpha = 1.0
    control_price_max_discount = {1: 0.2, 2: 0.3, 3: 0.3, 4: 0.0}
    tolerance_develop_rate = {
        1: 1.0 / 7.0,
        2: 2.5 / 7.0,
        3: 2.0 / 7.0,
        4: 0.5 / 7.0
    }
    control_price = {1: 1.0, 2: 0.8, 3: 0.7, 4: 0.5}
    # below two changes over stages according to x
    control_beta = {1: 2.0 / 7.0, 2: 3.0 / 7.0, 3: 3.0 / 7.0, 4: 5.0 / 7.0}

    payed_price_sum = 0
    above_threshold = 0

    init_pest_frac = np.random.RandomState(self._random_seed).beta(
        init_pest_frac_alpha, init_pest_frac_beta, size=(n_simulations,))
    curr_pest_frac = init_pest_frac
    for i in range(n_stages):
      spread_rate = np.random.RandomState(self._random_seed).beta(
          spread_alpha, spread_beta, size=(n_simulations,))
      do_control = x[i] > 0
      if do_control:
        control_rate = np.random.RandomState(self._random_seed).beta(
            control_alpha, control_beta[x[i]], size=(n_simulations,))
        next_pest_frac = self._pest_spread(curr_pest_frac, spread_rate,
                                           control_rate, True)
        # Tolerance has been developed for pesticide type 1.
        control_beta[x[i]] += tolerance_develop_rate[x[i]] / float(n_stages)
        # You will get a discount.
        payed_price = control_price[x[i]] * (
            1.0 - control_price_max_discount[x[i]] / float(n_stages) *
            float(np.sum(x == x[i])))
      else:
        next_pest_frac = self._pest_spread(curr_pest_frac, spread_rate, 0,
                                           False)
        payed_price = 0
      payed_price_sum += payed_price
      above_threshold += np.mean(curr_pest_frac > u)
      curr_pest_frac = next_pest_frac

    return payed_price_sum + above_threshold


# MAXSAT Files can be found in
# https://github.com/QUVA-Lab/COMBO/tree/master/COMBO/experiments/MaxSAT/maxsat2018_data.

MAXSAT28_FILE = 'maxsat2018_data/maxcut-johnson8-2-4.clq.wcnf'
MAXSAT43_FILE = 'maxsat2018_data/maxcut-hamming8-2.clq.wcnf'
MAXSAT60_FILE = 'maxsat2018_data/frb-frb10-6-4.wcnf'


class MAXSATExperimenter(experimenter.Experimenter):
  """MAXSAT Problem."""

  def __init__(self, data_filename: str):
    self._data_filename = data_filename
    f = fileOpen(self._data_filename, 'rt')
    line_str = f.readline()
    while line_str[:2] != 'p ':
      line_str = f.readline()
    self._n_variables = int(line_str.split(' ')[2])
    self._n_clauses = int(line_str.split(' ')[3])
    self._n_vertices = np.array([2] * self._n_variables)
    raw_clauses = [(float(clause_str.split(' ')[0]),
                    clause_str.split(' ')[1:-1])
                   for clause_str in f.readlines()]
    f.close()
    weights = np.array([elm[0] for elm in raw_clauses]).astype(np.float32)
    weight_mean = np.mean(weights)
    weight_std = np.std(weights)
    self._weights = (weights - weight_mean) / weight_std
    self._clauses = []
    for _, clause in raw_clauses:
      pair = ([abs(int(elm)) - 1 for elm in clause],
              [int(elm) > 0 for elm in clause])
      self._clauses.append(pair)

    self._problem_statement = self.problem_statement()

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    for suggestion in suggestions:
      # TODO: Switch to using StudyConfig.
      bools = [
          int(suggestion.parameters[f'x_{i}'].value == 'True')
          for i in range(self._n_variables)
      ]
      x = np.array(bools, dtype=bool)
      satisfied = np.array([
          (x[clause[0]] == clause[1]).any() for clause in self._clauses
      ])
      evaluation = -np.sum(self._weights * satisfied)
      suggestion.complete(
          pyvizier.Measurement(metrics={
              self._problem_statement.single_objective_metric_name: evaluation
          }))

  def problem_statement(self) -> pyvizier.ProblemStatement:
    problem_statement = pyvizier.ProblemStatement()
    root = problem_statement.search_space.root
    for i in range(self._n_variables):
      root.add_bool_param(name=f'x_{i}')
    problem_statement.metric_information.append(
        pyvizier.MetricInformation(
            name='main_objective', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE))
    return problem_statement



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Failing designers used for testing."""

from typing import Optional, Sequence
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import random


class FailedSuggestError(Exception):
  """Exception to raise during failing suggest call."""


class FailingDesigner(vza.Designer):
  """Failing Designer.

  The designer raises exception at every suggest call.
  """

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    pass

  def suggest(
      self, count: Optional[int] = None
  ) -> Sequence[vz.TrialSuggestion]:
    raise FailedSuggestError()


class AlternateFailingDesigner(vza.Designer):
  """Alternate Failing Designer.

  The designer raises exception at every second suggest call.

  Note: the designer doesn't persist a state across different instantiations and
  therefore should only be used during in memory run.
  """

  def __init__(self, search_space: vz.SearchSpace):
    self._suggest_count = 0
    self._random_designer = random.RandomDesigner(search_space)

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    pass

  def suggest(
      self, count: Optional[int] = None
  ) -> Sequence[vz.TrialSuggestion]:
    count = count or 1
    self._suggest_count += count
    if self._suggest_count % 2 == 0 or count > 1:
      raise FailedSuggestError()
    else:
      return self._random_designer.suggest(1)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for shifting_experimenter."""

import numpy as np
from vizier import pyvizier
from vizier._src.benchmarks.experimenters import numpy_experimenter
from vizier._src.benchmarks.experimenters import shifting_experimenter
from vizier._src.benchmarks.experimenters.synthetic import bbob

from absl.testing import absltest
from absl.testing import parameterized


class ShiftingExperimenterTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('Sphere', bbob.Sphere), ('Rastrigin', bbob.Rastrigin),
      ('BuecheRastrigin', bbob.BuecheRastrigin),
      ('LinearSlope', bbob.LinearSlope),
      ('AttractiveSector', bbob.AttractiveSector),
      ('StepEllipsoidal', bbob.StepEllipsoidal),
      ('RosenbrockRotated', bbob.RosenbrockRotated), ('Discus', bbob.Discus),
      ('BentCigar', bbob.BentCigar), ('SharpRidge', bbob.SharpRidge),
      ('DifferentPowers', bbob.DifferentPowers),
      ('Weierstrass', bbob.Weierstrass), ('SchaffersF7', bbob.SchaffersF7),
      ('SchaffersF7IllConditioned', bbob.SchaffersF7IllConditioned),
      ('GriewankRosenbrock', bbob.GriewankRosenbrock),
      ('Schwefel', bbob.Schwefel), ('Katsuura', bbob.Katsuura),
      ('Lunacek', bbob.Lunacek), ('Gallagher101Me', bbob.Gallagher101Me))
  def test_numpy_experimenter(self, func):
    dim = 2
    shift = 1.2
    exptr = numpy_experimenter.NumpyExperimenter(
        func, bbob.DefaultBBOBProblemStatement(dim))
    shifted_exptr = shifting_experimenter.ShiftingExperimenter(
        exptr=exptr, shift=np.asarray(shift))

    parameters = exptr.problem_statement().search_space.parameters
    self.assertLen(parameters, dim)

    t = pyvizier.Trial(parameters={
        param.name: float(index) for index, param in enumerate(parameters)
    })
    t_shifted = pyvizier.Trial(parameters={
        param.name: float(index) + shift
        for index, param in enumerate(parameters)
    })

    exptr.evaluate([t_shifted])
    shifted_exptr.evaluate([t])
    metric_name = exptr.problem_statement().metric_information.item().name

    self.assertAlmostEqual(
        t_shifted.final_measurement.metrics[metric_name].value,
        t.final_measurement.metrics[metric_name].value)
    self.assertEqual(t.status, t_shifted.status)

    # Check parameter bounds are shifted.
    shifted_parameters = shifted_exptr.problem_statement(
    ).search_space.parameters
    for param, shifted_param in zip(parameters, shifted_parameters):
      self.assertEqual(param.bounds[0], shifted_param.bounds[0])
      self.assertEqual(param.bounds[1] - shift, shifted_param.bounds[1])

  def test_evaluate_shift(self):
    dim = 2
    shift = np.array([1.2, -2.3])
    func = bbob.Sphere
    exptr = numpy_experimenter.NumpyExperimenter(
        func, bbob.DefaultBBOBProblemStatement(dim))
    shifted_exptr = shifting_experimenter.ShiftingExperimenter(
        exptr=exptr, shift=np.asarray(shift))

    parameters = exptr.problem_statement().search_space.parameters
    self.assertLen(parameters, dim)

    t = pyvizier.Trial(parameters={
        param.name: float(index) for index, param in enumerate(parameters)
    })
    t_shifted = pyvizier.Trial(
        parameters={
            param.name: float(index) + shift[index]
            for index, param in enumerate(parameters)
        })

    exptr.evaluate([t_shifted])
    shifted_exptr.evaluate([t])
    metric_name = exptr.problem_statement().metric_information.item().name

    self.assertAlmostEqual(
        t_shifted.final_measurement.metrics[metric_name].value,
        t.final_measurement.metrics[metric_name].value)
    self.assertEqual(t.status, t_shifted.status)
    self.assertNotEqual(t.parameters, t_shifted.parameters)

  def test_shift_backward(self):
    dim = 2
    shift = np.array([1.2, -2.3])
    func = bbob.Sphere
    exptr = numpy_experimenter.NumpyExperimenter(
        func, bbob.DefaultBBOBProblemStatement(dim))
    shifted_exptr = shifting_experimenter.ShiftingExperimenter(
        exptr=exptr, shift=np.asarray(shift))
    # Test shift within bounds
    trial = pyvizier.Trial(parameters={'x0': 3.0, 'x1': 1.0})
    shifted_exptr._offset([trial], shift=-shift)
    self.assertEqual(trial.parameters.as_dict(), {
        'x0': 3.0 - 1.2,
        'x1': 1.0 + 2.3
    })
    # Test shift in out of bounds
    trial = pyvizier.Trial(parameters={'x0': -5.0, 'x1': 5.0})
    shifted_exptr._offset([trial], shift=-shift)
    self.assertEqual(trial.parameters.as_dict(), {'x0': -5.0, 'x1': 5.0})

  def test_shift_forward(self):
    dim = 2
    shift = np.array([1.2, -2.3])
    func = bbob.Sphere
    exptr = numpy_experimenter.NumpyExperimenter(
        func, bbob.DefaultBBOBProblemStatement(dim))
    shifted_exptr = shifting_experimenter.ShiftingExperimenter(
        exptr=exptr, shift=np.asarray(shift))
    # Test shift within bounds
    trial = pyvizier.Trial(parameters={'x0': 3.0, 'x1': 1.0})
    shifted_exptr._offset([trial], shift=shift)
    self.assertEqual(trial.parameters.as_dict(), {
        'x0': 3.0 + 1.2,
        'x1': 1.0 - 2.3
    })
    # Test shift in out of bounds
    trial = pyvizier.Trial(parameters={'x0': 5.0, 'x1': -5.0})
    shifted_exptr._offset([trial], shift=shift)
    self.assertEqual(trial.parameters.as_dict(), {'x0': 5.0, 'x1': -5.0})

  @parameterized.parameters((True,), (False,))
  def test_shift_forward_oob(self, should_restrict):
    dim = 2
    shift = np.array([2.2, -2.3])
    func = bbob.Sphere
    exptr = numpy_experimenter.NumpyExperimenter(
        func, bbob.DefaultBBOBProblemStatement(dim)
    )
    shifted_exptr = shifting_experimenter.ShiftingExperimenter(
        exptr=exptr, shift=np.asarray(shift), should_restrict=should_restrict
    )
    # Test OOB shifts does not change parameter values.
    trial = pyvizier.Trial(parameters={'x0': 3.0, 'x1': 1.0})
    shifted_exptr.evaluate([trial])
    self.assertEqual(trial.parameters.as_dict(), {'x0': 3.0, 'x1': 1.0})

    # Test OOB shifts stay within bounds.
    shifted_exptr._offset([trial], shift=shift)
    # x0 is shifted OOB, so clip at 5.0.
    self.assertEqual(trial.parameters.as_dict(), {'x0': 5.0, 'x1': 1.0 - 2.3})

  def test_large_shift(self):
    dim = 2
    shift = np.array([10.2, 20.3])
    func = bbob.Sphere
    exptr = numpy_experimenter.NumpyExperimenter(
        func, bbob.DefaultBBOBProblemStatement(dim))

    with self.assertRaisesRegex(ValueError, 'is too large'):
      shifting_experimenter.ShiftingExperimenter(
          exptr=exptr, shift=np.asarray(shift))

    shifted_exptr = shifting_experimenter.ShiftingExperimenter(
        exptr=exptr, shift=np.asarray(shift), should_restrict=False
    )
    parameters = exptr.problem_statement().search_space.parameters
    self.assertEqual(
        parameters, shifted_exptr.problem_statement().search_space.parameters
    )
    t = pyvizier.Trial(
        parameters={
            param.name: float(index) for index, param in enumerate(parameters)
        }
    )
    shifted_exptr.evaluate([t])

if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for serialization."""

import numpy as np
from vizier._src.algorithms.designers.eagle_strategy import serialization
from vizier._src.algorithms.designers.eagle_strategy import testing

from absl.testing import absltest


class SerializationTest(absltest.TestCase):

  def test_restore_pool(self):
    firefly_pool = testing.create_fake_populated_firefly_pool(capacity=20)
    encoded = serialization.partially_serialize_firefly_pool(firefly_pool)
    # Restore the firefly pool.
    utils = firefly_pool.utils
    restored_firefly_pool = serialization.restore_firefly_pool(utils, encoded)
    # Check that the restored and original firefly_pool are the same.
    self.assertEqual(restored_firefly_pool.capacity, firefly_pool.capacity)
    self.assertEqual(restored_firefly_pool._last_id, firefly_pool._last_id)
    self.assertEqual(restored_firefly_pool._max_fly_id,
                     firefly_pool._max_fly_id)
    self.assertEqual(
        set(firefly_pool._pool.keys()), set(restored_firefly_pool._pool.keys()))
    for fly_id, firefly in firefly_pool._pool.items():
      restored_firefly = restored_firefly_pool._pool[fly_id]
      self.assertEqual(restored_firefly.id_, firefly.id_)
      self.assertEqual(restored_firefly.perturbation, firefly.perturbation)
      self.assertEqual(restored_firefly.generation, firefly.generation)
      self.assertEqual(restored_firefly.trial.parameters,
                       firefly.trial.parameters)
      self.assertEqual(
          restored_firefly.trial.final_measurement.metrics['objective'].value,
          firefly.trial.final_measurement.metrics['objective'].value)

  def test_restore_rng(self):
    rng = np.random.default_rng(0)
    serialized_rng = serialization.serialize_rng(rng)
    rand_value = rng.normal()
    restored_rng = serialization.restore_rng(serialized_rng)
    assert restored_rng.normal() == rand_value


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for Sklearn classifier."""

import numpy as np
from sklearn import svm
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.gaussian_process.kernels import ConstantKernel
from sklearn.gaussian_process.kernels import RBF
from sklearn.metrics.pairwise import euclidean_distances
from vizier._src.algorithms.classification import classifiers

from absl.testing import absltest
from absl.testing import parameterized

svm_classifier = svm.SVC(kernel='rbf', C=1.)
gpc_classifier = GaussianProcessClassifier(
    kernel=ConstantKernel(1.) * RBF(length_scale=1.))


# TODO: convert it into a generic ClassifierTest for all subclasses.
class SklearnClassifierTest(parameterized.TestCase):
  """Class for testing the classifier."""

  def setUp(self):
    super().setUp()
    self.features_train = np.array([[1.], [2.], [3.], [4.]])
    self.labels_train = np.array([0, 0, 1, 1])
    self.labels_train_invalid = np.array([10., 10., 11., 11.])
    self.features_test = np.array([[-1.], [6.], [0.], [6.]])

  def test_raise_error_invalid_labels(self):
    classifier_instance = classifiers.SklearnClassifier(
        classifier=gpc_classifier,
        features=self.features_train,
        labels=self.labels_train_invalid,
        features_test=self.features_test,
        eval_metric='probability')
    with self.assertRaises(ValueError):
      classifier_instance._check_labels_values()

  @parameterized.parameters([
      dict(classifier=svm_classifier, eval_metric='decision'),
      dict(classifier=gpc_classifier, eval_metric='probability')
  ])
  def test_scores_shape(self, classifier, eval_metric):
    classifier_instance = classifiers.SklearnClassifier(
        classifier=classifier,
        features=self.features_train,
        labels=self.labels_train,
        features_test=self.features_test,
        eval_metric=eval_metric)
    scores = classifier_instance()
    self.assertEqual(scores.shape[0], self.features_test.shape[0])

  @parameterized.parameters([
      dict(classifier=svm_classifier, eval_metric='decision', threshold=0.),
      dict(classifier=gpc_classifier, eval_metric='probability', threshold=0.5)
  ])
  def test_labels_shape(self, classifier, eval_metric, threshold):
    classifier_instance = classifiers.SklearnClassifier(
        classifier=classifier,
        features=self.features_train,
        labels=self.labels_train,
        features_test=self.features_test,
        eval_metric=eval_metric)
    scores = classifier_instance()
    labels_test_pred = (scores >= threshold).astype(float)
    labels_test_real = np.array([0, 1, 0, 1])
    self.assertTrue((labels_test_pred == labels_test_real).all())

  @parameterized.parameters([dict(label_val=0.), dict(label_val=1.)])
  def test_raise_error_identical_labels(self, label_val):
    labels_train_identical = np.ones(
        (self.features_train.shape[0],)) * label_val
    classifier_instance = classifiers.SklearnClassifier(
        classifier=gpc_classifier,
        features=self.features_train,
        labels=labels_train_identical,
        features_test=self.features_test,
        eval_metric='probability')
    with self.assertRaises(ValueError):
      classifier_instance._check_labels_values()

  @parameterized.parameters([
      dict(classifier=svm_classifier, eval_metric='decision'),
      dict(classifier=gpc_classifier, eval_metric='probability')
  ])
  def test_scores_range(self, classifier, eval_metric):
    classifier_instance = classifiers.SklearnClassifier(
        classifier=classifier,
        features=self.features_train,
        labels=self.labels_train,
        features_test=self.features_test,
        eval_metric=eval_metric)
    scores = classifier_instance()
    if eval_metric == 'probability':
      self.assertGreaterEqual(scores.min(), 0.)
      self.assertLessEqual(scores.max(), 1.)
    else:
      max_distance = max(
          euclidean_distances(self.features_test).max(),
          euclidean_distances(self.features_train).max(),
          euclidean_distances(self.features_train, self.features_test).max())
      self.assertLessEqual(scores.max(), max_distance)

  @parameterized.parameters([
      dict(classifier=svm_classifier, eval_metric='decision', threshold=0.),
      dict(classifier=gpc_classifier, eval_metric='probability', threshold=0.5)
  ])
  def test_prediction_on_train_data(self, classifier, eval_metric, threshold):
    classifier_instance = classifiers.SklearnClassifier(
        classifier=classifier,
        features=self.features_train,
        labels=self.labels_train,
        features_test=self.features_train,
        eval_metric=eval_metric)
    scores = classifier_instance()
    labels_test_pred = (scores >= threshold).astype(float)
    self.assertTrue((labels_test_pred == self.labels_train).all())


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import jax.numpy as jnp
from vizier import benchmarks
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import random
from vizier._src.algorithms.designers import scalarization
from vizier._src.algorithms.designers import scalarizing_designer
from vizier._src.algorithms.designers.eagle_strategy import eagle_strategy
from vizier._src.algorithms.ensemble import ensemble_design
from vizier._src.algorithms.ensemble import ensemble_designer
from vizier.benchmarks import experimenters

from absl.testing import absltest
from absl.testing import parameterized


class EnsembleDesignerTest(parameterized.TestCase):

  # pylint: disable=g-long-lambda
  @parameterized.parameters(
      (ensemble_design.EXP3UniformEnsembleDesign),
      (ensemble_design.EXP3IXEnsembleDesign),
      (
          lambda ind: ensemble_design.AdaptiveEnsembleDesign(
              indices=ind, max_lengths=[50, 100, 200]
          )
      ),
  )
  def testSmartEnsemblingWithGenerator(self, ensemble_design_factory):
    dim = 6
    bbob_factory = experimenters.BBOBExperimenterFactory('Sphere', dim)
    exptr_factory = experimenters.SingleObjectiveExperimenterFactory(
        bbob_factory
    )

    def ensemble_designer_factory(config: vz.ProblemStatement, seed: int):
      random_designer = random.RandomDesigner(config.search_space, seed=seed)
      eagle = eagle_strategy.EagleStrategyDesigner(config, seed=seed)
      reward_generator = ensemble_designer.ObjectiveRewardGenerator(
          config, reward_regularization=0.1
      )
      return ensemble_designer.EnsembleDesigner(
          {'random': random_designer, 'eagle': eagle},
          ensemble_design_factory=ensemble_design_factory,
          reward_generator=reward_generator,
      )

    benchmark_state_factory = (
        benchmarks.ExperimenterDesignerBenchmarkStateFactory(
            designer_factory=ensemble_designer_factory,
            experimenter_factory=exptr_factory,
        )
    )
    bench_state = benchmark_state_factory()
    runner = benchmarks.BenchmarkRunner(
        benchmark_subroutines=[benchmarks.GenerateAndEvaluate(5)],
        num_repeats=50,
    )
    runner.run(bench_state)
    self.assertEmpty(
        bench_state.algorithm.supporter.GetTrials(
            status_matches=vz.TrialStatus.ACTIVE
        )
    )
    all_trials = bench_state.algorithm.supporter.GetTrials()
    self.assertLen(all_trials, 250)

    num_random_trials = 0
    num_eagle_trials = 0
    for t in all_trials:
      if (
          t.metadata.ns(ensemble_designer.ENSEMBLE_NS).get(
              ensemble_designer.EXPERT_KEY
          )
          == 'random'
      ):
        num_random_trials += 1
      elif (
          t.metadata.ns(ensemble_designer.ENSEMBLE_NS).get(
              ensemble_designer.EXPERT_KEY
          )
          == 'eagle'
      ):
        num_eagle_trials += 1
    self.assertEqual(num_random_trials + num_eagle_trials, 250)

  @parameterized.parameters(
      (ensemble_design.EXP3UniformEnsembleDesign),
      (ensemble_design.EXP3IXEnsembleDesign),
      (
          lambda ind: ensemble_design.AdaptiveEnsembleDesign(
              indices=ind, max_lengths=[50, 100, 200]
          )
      ),
  )
  def testPendingMulitobjectiveUpdate(self, ensemble_design_factory):
    dim = 2
    func1 = experimenters.bbob.Sphere
    func2 = experimenters.bbob.Rastrigin
    exptr1 = experimenters.NumpyExperimenter(
        func1, experimenters.bbob.DefaultBBOBProblemStatement(dim)
    )
    exptr2 = experimenters.NumpyExperimenter(
        func2, experimenters.bbob.DefaultBBOBProblemStatement(dim)
    )
    exptr = experimenters.MultiObjectiveExperimenter(
        {'m1': exptr1, 'm2': exptr2}
    )

    def ensemble_designer_factory(config: vz.ProblemStatement, seed: int):
      random_designer = random.RandomDesigner(config.search_space, seed=seed)

      def eagle_designer_factory(ps, seed):
        return eagle_strategy.EagleStrategyDesigner(
            problem_statement=ps, seed=seed
        )

      scalarized_eagle = scalarizing_designer.ScalarizingDesigner(
          config,
          eagle_designer_factory,
          scalarizer=scalarization.HyperVolumeScalarization(
              weights=jnp.ones(len(config.metric_information))
          ),
      )

      reward_generator = ensemble_designer.ObjectiveRewardGenerator(
          config, reward_regularization=0.1
      )
      return ensemble_designer.EnsembleDesigner(
          {'random': random_designer, 'eagle': scalarized_eagle},
          ensemble_design_factory=ensemble_design_factory,
          reward_generator=reward_generator,
      )

    benchmark_state_factory = benchmarks.DesignerBenchmarkStateFactory(
        designer_factory=ensemble_designer_factory,
        experimenter=exptr,
    )
    bench_state = benchmark_state_factory()
    runner = benchmarks.BenchmarkRunner(
        benchmark_subroutines=[benchmarks.GenerateSuggestions(1)],
        num_repeats=5,
    )
    runner.run(bench_state)
    self.assertLen(
        bench_state.algorithm.supporter.GetTrials(
            status_matches=vz.TrialStatus.ACTIVE
        ),
        5,
    )


if __name__ == '__main__':
  absltest.main()
